#include "discamb/IO/xd_io.h"
#include "discamb/BasicUtilities/Exception.h"
#include "discamb/MathUtilities/MathUtilities.h"
#include "discamb/BasicUtilities/StringUtilities.h"
#include "discamb/CrystalStructure/ReciprocalLatticeUnitCell.h"
#include "discamb/Scattering/HansenCoppensStructureFactorCalculator.h"
#include "discamb/MathUtilities/RealSphericalHarmonics.h"
#include "discamb/Scattering/SlaterTypeOrbitalScattering.h"
#include "discamb/Scattering/multipole_scattering.h"



#include <iostream>
#include <iomanip>

using namespace std;
using namespace discamb;

/* Lets represent some deformation-valence electron density as the following sum:
   
   rho_def_val(r) = R_1(r) * Y_10(theta,phi) + R_2(r) * Y_20(theta,phi)
   
   where R_i are radial functions and Y_lm are density normalized spherical harmonics.
   Lets choose:
   
   R_1(r) = a_1 * r * exp(-b_1*r) + a_2 * r^2 * exp(-b_2*r) + a_3 * r^3 * exp(-b_3*r)
   R_2(r) = a_4 * r^3 * exp(-b_4*r)

*/

// the parameters of the deformation valence (as in the comment/equation above)
const double a[] = { 0.023, 0.126, 0.078, 0.078 };
const double b[] = { 1.023, 1.706, 2.082, 1.924 };
const size_t powerR[] = { 1, 2, 3, 3 };


// the values assigned to the parameters have no physical sense
// they are used only to illustrate the idea


void calcScatteringFactorsApproach1(const vector<Vector3i> &hkl, const UnitCell &unitCell, vector<complex<double> > &scatteringFactors);
void calcScatteringFactorsApproach2(const vector<Vector3i> &hkl, const UnitCell &unitCell, vector<complex<double> > &scatteringFactors);
void calcScatteringFactorsApproach3(const vector<Vector3i> &hkl, const UnitCell &unitCell, vector<complex<double> > &scatteringFactors);
void calcScatteringFactorsApproach4(const vector<Vector3i> &hkl, const UnitCell &unitCell, vector<complex<double> > &scatteringFactors);

void run()
{
    size_t approachIdx, nApproaches=3, hklIdx, nHkl = 3;
    vector<Vector3i> hkl_indices(nHkl);
    UnitCell unitCell(7.63,4.28,9.72,90,90,90);
    vector<vector<complex<double> > > scatteringFactors(nApproaches);
    hkl_indices[0].set(0, 0, 1);
    hkl_indices[1].set(1, 1, 0);
    hkl_indices[2].set(2, 2, 2);

    calcScatteringFactorsApproach1(hkl_indices, unitCell, scatteringFactors[0]);
    calcScatteringFactorsApproach2(hkl_indices, unitCell, scatteringFactors[1]);
    calcScatteringFactorsApproach3(hkl_indices, unitCell, scatteringFactors[2]);
    
    // print out results

    cout<< endl << "   H   K    L";
    for(approachIdx=0;approachIdx<nApproaches;approachIdx++)
        cout<< setw(28) << string("   approach ") + string_utilities::convertToString(approachIdx+1) +string("     ");
    cout<< endl << endl;

    for(hklIdx=0;hklIdx<nHkl; hklIdx++)
    {
        cout << setw(4) << hkl_indices[hklIdx](0) 
             << setw(4) << hkl_indices[hklIdx](1) 
             << setw(4) << hkl_indices[hklIdx](2);
        for (approachIdx = 0; approachIdx<nApproaches; approachIdx++)
            cout << setw(14) << setprecision(6) << scatteringFactors[approachIdx][hklIdx].real() 
                 << setw(14) << setprecision(6) << scatteringFactors[approachIdx][hklIdx].imag();
        cout<< endl;
    }

}

int main()
{
    
    try
    {
        run();
    }
    catch (const Exception & e)
    {
        cout<< e.what();
    }
    catch (const std::exception& e)
    {
        cout<< "standad exception " << e.what() << endl;
    }
    
    return 0;
}


// the four terms corresponds to four atoms
void calcScatteringFactorsApproach1(
    const vector<Vector3i> &hkl, 
    const UnitCell &unitCell,
    vector<complex<double> > &scatteringFactors)
{
    size_t nTerms=4;
    
    // set wfn types
    
    vector<HC_WfnType> wfnTypeParameters(nTerms);
    wfnTypeParameters[0].deformation_valence_power.resize(2, 1);
    wfnTypeParameters[1].deformation_valence_power.resize(2, 2);
    wfnTypeParameters[2].deformation_valence_power.resize(2, 3);
    wfnTypeParameters[3].deformation_valence_power.resize(3, 3);

    for (size_t i = 0; i<nTerms; i++)
        wfnTypeParameters[i].deformation_valence_exponent = b[i];

    // set atom type parameters

    vector<HC_AtomTypeParameters> atomTypeParameters(4);

    // set all Plms to 0
    for (size_t i = 0; i<nTerms; i++)
    {
        atomTypeParameters[i].p_lm.resize(3);
        for (size_t l = 0; l<3; l++)
            atomTypeParameters[i].p_lm[l].assign(2 * l + 1, 0.0);
    }

    // assign some Plms to a[i]
    for (int i = 0; i<3; i++)
        atomTypeParameters[i].p_lm[1][1] = a[i]/ sto_atomic_wfn::stoDensityNormalizationFactor(powerR[i],b[i]);
    atomTypeParameters[3].p_lm[2][2] = a[3] / sto_atomic_wfn::stoDensityNormalizationFactor(powerR[3], b[3]);
 
 

    HC_ModelParameters hcModelParameters;

    hcModelParameters.atom_to_type_map.resize(nTerms);
    hcModelParameters.atom_to_wfn_map.resize(nTerms);
    for (size_t i = 0; i<nTerms; i++)
    {
        hcModelParameters.atom_to_type_map[i] = i;
        hcModelParameters.atom_to_wfn_map[i] = i;
    }

    hcModelParameters.type_parameters = atomTypeParameters;
    hcModelParameters.wfn_parameters = wfnTypeParameters;

    Crystal crystal;
    vector<Matrix3d> lcs(nTerms);
    

    crystal.unitCell = unitCell;

    crystal.atoms.resize(nTerms);
    for( auto &atom: crystal.atoms)
    {
        atom.occupancy=1.0; 
        atom.multiplicity=1; 
    }


    for(auto & localCoordinateSystem : lcs)
        localCoordinateSystem.setToIdentity();

    HansenCoppensStructureFactorCalculator calculator(crystal, hcModelParameters);


    calculator.calculateStructureFactors(crystal.atoms,lcs,hkl,scatteringFactors);
}


void calcScatteringFactorsApproach2(
    const vector<Vector3i> &hkl, 
    const UnitCell &unitCell, 
    vector<complex<double> > &scatteringFactors)
{
    size_t hklIdx,nHkl = hkl.size();
    vector<Vector3d> hkl_cartesian(nHkl);
    Vector3d normalizedHklVector;
    ReciprocalLatticeUnitCell reciprocalLattice(unitCell);
    const complex<double> imaginary_unit(0,1);
    double hVectorLength;
    //double y_10; // real spherical harmonics Y_{1,0} and Y_{2,0}

    for(hklIdx=0;hklIdx<nHkl;hklIdx++)
        reciprocalLattice.fractionalToCartesian(hkl[hklIdx],hkl_cartesian[hklIdx]);

    scatteringFactors.assign(nHkl,complex<double>(0.0));

    for ( hklIdx = 0 ; hklIdx < nHkl ; hklIdx++ )
    {
        // calculate length of hkl vector and 
        // normalized version of the vector 
        hVectorLength = sqrt(hkl_cartesian[hklIdx]*hkl_cartesian[hklIdx]);
        normalizedHklVector = hkl_cartesian[hklIdx] / hVectorLength;

        // in the case of the hkl vector = (0,0,0) there is no contribution 
        // from multipolar terms with l>0

        if(hVectorLength == 0.0)
        {
            scatteringFactors[hklIdx]=0;
            continue;
        }

        // 

        scatteringFactors[hklIdx] = 
            // term 1
                                        a[0]*4*M_PI*imaginary_unit*sto_scattering::gFunction<1>(3,hVectorLength,b[0])*
                                        real_spherical_harmonics::densityNormalized(normalizedHklVector,1,0) 
            //term 2
                                    +   a[1] * 4 * M_PI*imaginary_unit*sto_scattering::gFunction<1>(4, hVectorLength, b[1])*
                                        real_spherical_harmonics::densityNormalized(normalizedHklVector,1,0) 
            //term 3
                                    +   a[2] * 4 * M_PI*imaginary_unit*sto_scattering::gFunction<1>(5, hVectorLength, b[2])*
                                        real_spherical_harmonics::densityNormalized(normalizedHklVector,1,0) 
            //term 4
                                    -   a[3] * 4 * M_PI*sto_scattering::gFunction<2>(5, hVectorLength, b[3])*
                                        real_spherical_harmonics::densityNormalized(normalizedHklVector,2,0);
    }
}


void calcScatteringFactorsApproach3(
    const vector<Vector3i> &hkl,
    const UnitCell &unitCell,
    vector<complex<double> > &scatteringFactors)
{
    size_t hklIdx, nHkl = hkl.size();
    Vector3d hkl_cartesian, hkl_normalized;
    double hVectorLength;
    double radialScatteringFactor1, radialScatteringFactor2;
    complex<double> angularScatteringFactor1, angularScatteringFactor2;
    ReciprocalLatticeUnitCell reciprocalLattice(unitCell);

    
        

    scatteringFactors.resize(nHkl);

    for (hklIdx = 0; hklIdx < nHkl; hklIdx++)
    {
        // converts from Miller indices to vector in Cartesian coordinates
        reciprocalLattice.fractionalToCartesian(hkl[hklIdx], hkl_cartesian);
        hVectorLength = sqrt(hkl_cartesian * hkl_cartesian);
        hkl_normalized = hkl_cartesian / hVectorLength;

        // in the case of the hkl vector = (0,0,0) there is no contribution 
        // from multipolar terms with l>0

        if (hkl[hklIdx]== Vector3i(0,0,0))
        {
            scatteringFactors[hklIdx] = 0;
            continue;
        }

        // 

        radialScatteringFactor1 = a[0] * sto_scattering::gFunction<1>(powerR[0] + 2, hVectorLength, b[0]) +
                                  a[1] * sto_scattering::gFunction<1>(powerR[1] + 2, hVectorLength, b[1]) +
                                  a[2] * sto_scattering::gFunction<1>(powerR[2] + 2, hVectorLength, b[2]);
        
        radialScatteringFactor2 = a[3] * sto_scattering::gFunction<2>(powerR[3] + 2, hVectorLength, b[3]);

        angularScatteringFactor1 = multipole_scattering::multipoleScattering(1, 0, hkl_normalized, true);
        angularScatteringFactor2 = multipole_scattering::multipoleScattering(2, 0, hkl_normalized, true);
        

        scatteringFactors[hklIdx] = radialScatteringFactor1 * angularScatteringFactor1 +
                                    radialScatteringFactor2 * angularScatteringFactor2;
    }
}

void calcScatteringFactorsApproach4(
    const vector<Vector3i> &hkl,
    const UnitCell &unitCell,
    vector<complex<double> > &scatteringFactors)
{
    size_t hklIdx, nHkl = hkl.size();
    Vector3d hkl_cartesian, hkl_normalized;
    double hVectorLength;
    double radialScatteringFactor1, radialScatteringFactor2;
    complex<double> angularScatteringFactor1, angularScatteringFactor2;
    ReciprocalLatticeUnitCell reciprocalLattice(unitCell);




    scatteringFactors.resize(nHkl);

    for (hklIdx = 0; hklIdx < nHkl; hklIdx++)
    {
        // converts from Miller indices to vector in Cartesian coordinates
        reciprocalLattice.fractionalToCartesian(hkl[hklIdx], hkl_cartesian);
        hVectorLength = sqrt(hkl_cartesian * hkl_cartesian);
        hkl_normalized = hkl_cartesian / hVectorLength;

        // in the case of the hkl vector = (0,0,0) there is no contribution 
        // from multipolar terms with l>0

        if (hkl[hklIdx] == Vector3i(0, 0, 0))
        {
            scatteringFactors[hklIdx] = 0;
            continue;
        }

        // 

        radialScatteringFactor1 = a[0] * sto_scattering::gFunction<1>(powerR[0] + 2, hVectorLength, b[0]) +
            a[1] * sto_scattering::gFunction<1>(powerR[1] + 2, hVectorLength, b[1]) +
            a[2] * sto_scattering::gFunction<1>(powerR[2] + 2, hVectorLength, b[2]);

        radialScatteringFactor2 = a[3] * sto_scattering::gFunction<2>(powerR[3] + 2, hVectorLength, b[3]);

        angularScatteringFactor1 = multipole_scattering::multipoleScattering(1, 0, hkl_normalized, true);
        angularScatteringFactor2 = multipole_scattering::multipoleScattering(2, 0, hkl_normalized, true);


        scatteringFactors[hklIdx] = radialScatteringFactor1 * angularScatteringFactor1 +
            radialScatteringFactor2 * angularScatteringFactor2;
    }
}
