#include "Target.h"
#include "discamb/BasicUtilities/StringUtilities.h"

using namespace std;
using namespace discamb;

namespace{

    void getParameterFlags(
        const std::string &line,
        vector<bool> &flags)
    {
        string one("1");
        vector<string> words;
        size_t nWords;

        string_utilities::split(line, words);
        nWords = words.size();
        flags.resize(nWords);

        for(size_t i=0;i<nWords;i++)
            flags[i] = (words[i] == one);
    }

}

Target::Target(
    HansenCoppensStructureFactorCalculator *engine, 
    const Crystal &crystal,
    const std::vector<XdLocalCoordinateSystem> &lcs)
{
    mA = mB = 0.0;
    const vector<AtomInCrystal> &atoms = crystal.atoms;
    mEngine = engine;
    mCrystal = crystal;

    size_t nAtoms, atomIdx, i, nAdps, parameterIdx;
    nAtoms = atoms.size();

    mOcc2Param.resize(nAtoms);
    mXyz2Param.resize(nAtoms,vector<int>(3));
    mAdp2Param.resize(nAtoms);

    parameterIdx = 0;

    for(atomIdx=0; atomIdx<nAtoms; atomIdx++)
    {
        if(atoms[atomIdx].occupancy != 1)
            mOcc2Param[atomIdx] = parameterIdx++;
        else
            mOcc2Param[atomIdx] = -1;

        for(i=0;i<3;i++)
            mXyz2Param[atomIdx][i] = parameterIdx++;
        
        nAdps = atoms[atomIdx].adp.size();
        mAdp2Param[atomIdx].resize(nAdps);

        for (i = 0; i<nAdps; i++)
            mAdp2Param[atomIdx][i] = parameterIdx++;

    }

    mN_Param = parameterIdx;

    mLcs = lcs;
    mLcsMatrices.resize(lcs.size());

    updateLcs(crystal);
}

Target::Target(
    HansenCoppensStructureFactorCalculator *engine,
    const Crystal &crystal,
    const std::string &optimizationOptionsFile,
    const std::vector<XdLocalCoordinateSystem> &lcs)
{
    initialize(engine, crystal, lcs, optimizationOptionsFile);
}


void Target::initialize(
    HansenCoppensStructureFactorCalculator *engine,
    const Crystal &crystal,
    const std::vector<XdLocalCoordinateSystem> &lcs, 
    const std::string &optimizationOptionsFile)
{
    mA = mB = 0.0;
    mEngine = engine;
    mCrystal = crystal;

    size_t nAtoms = crystal.atoms.size();

    mOcc2Param.resize(nAtoms);
    mXyz2Param.resize(nAtoms, vector<int>(3));
    mAdp2Param.resize(nAtoms);


    if(optimizationOptionsFile.empty())
        mapParameters(crystal);
    else
        mapParameters(optimizationOptionsFile);

    mLcs = lcs;
    mLcsMatrices.resize(lcs.size());

    updateLcs(crystal);
}

void Target::mapParameters(
    const Crystal &crystal)
{
    const vector<AtomInCrystal> &atoms = crystal.atoms;

    size_t nAtoms, atomIdx, i, nAdps, parameterIdx;
    nAtoms = atoms.size();

    parameterIdx = 0;

    for (atomIdx = 0; atomIdx<nAtoms; atomIdx++)
    {
        if (atoms[atomIdx].occupancy != 1)
            mOcc2Param[atomIdx] = parameterIdx++;
        else
            mOcc2Param[atomIdx] = -1;

        for (i = 0; i<3; i++)
            mXyz2Param[atomIdx][i] = parameterIdx++;

        nAdps = atoms[atomIdx].adp.size();
        mAdp2Param[atomIdx].resize(nAdps);

        for (i = 0; i<nAdps; i++)
            mAdp2Param[atomIdx][i] = parameterIdx++;

    }

    mN_Param = parameterIdx;

}

void Target::mapParameters(
    const std::string &optimizationOptionsFile)
{
    ifstream in(optimizationOptionsFile.c_str());
    string line;
    vector<bool> flags;
    size_t i, nAdps, nAtomParams, paramIdx;

    if(!in.good())
        on_error::throwException("can not read file with optimization options", __FILE__ , __LINE__ );

    size_t atomIdx=0;
    paramIdx = 0;

    while(in.good())
    {
        getline(in,line);
        getParameterFlags(line, flags);
        nAtomParams = flags.size();
        if(nAtomParams >= 4)
        {
            if(flags[0])
                mOcc2Param[atomIdx] = paramIdx++;
            else
                mOcc2Param[atomIdx] = -1;
            
            for(i=0;i<3;i++)
                if (flags[i+1])
                    mXyz2Param[atomIdx][i] = paramIdx++;
                else
                    mXyz2Param[atomIdx][i] = -1;

            nAdps = nAtomParams-4;
            mAdp2Param[atomIdx].resize(nAdps);
            for(i=0;i<nAdps;i++)
                if (flags[i + 4])
                    mAdp2Param[atomIdx][i] = paramIdx++;
                else
                    mAdp2Param[atomIdx][i] = -1;
            
            atomIdx++;
        }
    }

    in.close();

    mN_Param = paramIdx;

}


void Target::getParameters(
    std::vector<double> &parameters, 
    const Crystal &crystal)
{
    size_t i, j, nAdp, nAtoms = crystal.atoms.size();
    const vector<AtomInCrystal> &atoms = crystal.atoms;
    parameters.resize(mN_Param);
    
    for( i=0 ; i<nAtoms ; i++)
    {
        if(mOcc2Param[i] != -1)
            parameters[mOcc2Param[i]] = atoms[i].occupancy;
        
        for(j=0;j<3;j++)
            if(mXyz2Param[i][j] != -1)
            parameters[mXyz2Param[i][j]] = atoms[i].coordinates[j];

        nAdp = atoms[i].adp.size();

        for(j=0;j<nAdp;j++)
            if(mAdp2Param[i][j] != -1)
                parameters[mAdp2Param[i][j]] = atoms[i].adp[j];
    }
}

void Target::calculateWeights(
    std::vector<double> &weights)
{
    /*
    w = 1 / [sigma2(Fo2) + (aP)2 + bP]
        where P = [2Fc2 + Max(Fo2, 0)] / 3
        */

    size_t hklIdx, nHkl = mHkl.size();
    double i_calc, p;
    bool sf_calc_exist = !mF_calc.empty();

    weights.resize(nHkl);

    for(hklIdx=0; hklIdx<nHkl; hklIdx++)
    {
        i_calc = ( sf_calc_exist ? norm(mF_calc[hklIdx]) : 0 );
        p = ( 2 * i_calc + std::max( mObservedIntensities[hklIdx], 0.0 ) ) / 3.0;
        weights[hklIdx] = 1.0 / ( mIntensitiesSigma[hklIdx] * mIntensitiesSigma[hklIdx] + mA * mA * p * p + mB * p);
    }

}

Target::~Target()
{
}

void Target::setObservationsAndInit(
    const std::vector<Vector3i> hkl, 
    const std::vector<double> &intensities, 
    const std::vector<double> &sigma)
{
    mObservedIntensities = intensities;
    mIntensitiesSigma = sigma;
    mHkl = hkl;

    calculateSF(mF_calc);
    calculateWeights(mWeights);
    mScaleFactor = calculateScaleFactor();

}

double Target::calculateScaleFactor()
{
    double oc, cc, ic;

    oc = cc = 0.0;

    size_t nHkl = mHkl.size();
    for(size_t i=0; i<nHkl;i++)
    {
        ic = norm(mF_calc[i]);
        oc += mWeights[i] * mObservedIntensities[i] * ic;
        cc += mWeights[i] * ic * ic;
    }

    return oc/cc;
}

/** Calculate structure factors using the last parameterization of mCrystal*/

void Target::calculateSF(std::vector<std::complex<double> > &sf)
{
    updateLcs(mCrystal);
    mEngine->calculateStructureFactors(mCrystal.atoms, mLcsMatrices, mHkl, sf);
}

// sum over hkl weight[i]*I_obs[i]*I_obs[i] - the denominator in wR2
double Target::calc_wR2_denominator()
{
    
    double result = 0;
    size_t nHkl = mHkl.size();
    for (size_t i = 0; i<nHkl; i++)
        result += mWeights[i] * mObservedIntensities[i] * mObservedIntensities[i];
    return result;
}

double Target::calc_wR2_numerator()
{
    double diff, result = 0;
    size_t nHkl = mHkl.size();
    
    for (size_t i = 0; i<nHkl; i++)
    {
        diff = mObservedIntensities[i] - mScaleFactor * norm(mF_calc[i]);
        result += mWeights[i] * diff * diff;
    }
    return result;
}

double Target::calc_wR2_square()
{
    return calc_wR2_numerator() / calc_wR2_denominator();
}

void Target::calculate_dTargetdF(
    vector<complex<double> > &dTargetdF)
{
    size_t nHkl = mHkl.size();
    dTargetdF.resize(nHkl);
    double diff, denominator = calc_wR2_denominator();

    for(size_t i = 0 ; i < nHkl ; i++ )
    {
        diff = (mObservedIntensities[i] - mScaleFactor*norm(mF_calc[i]));
        dTargetdF[i] = -4 * mWeights[i] * diff * mF_calc[i] * mScaleFactor / denominator;
    }
    
}

double Target::targetAndDeriveatives(
    const std::vector<double> &parameters, 
    std::vector<double> &derivatives)
{
    vector<complex<double> > dTarget_dF;
    size_t i, j, nAdps, nAtoms = mCrystal.atoms.size();
    vector<bool> countAtomContribution(nAtoms ,true);
    parameterizeCrystal(parameters, mCrystal);
    calculateSF(mF_calc);

    calculateWeights(mWeights);
    mScaleFactor = calculateScaleFactor();
    calculate_dTargetdF(dTarget_dF);

    mEngine->calculateStructureFactorsAndDerivatives(
        mCrystal.atoms, mLcsMatrices, mHkl, mF_calc, 
        mD_target_dParam, dTarget_dF, countAtomContribution);
    
    derivatives.resize(mN_Param);

    for(i=0; i< nAtoms ; i++)
    {
        if( mOcc2Param[i] != -1)
            derivatives[ mOcc2Param[i] ] = mD_target_dParam[i].occupancy_derivatives;
        
        for( j=0; j<3; j++)
            if(mXyz2Param[i][j] != -1)
                derivatives[ mXyz2Param[i][j] ] = mD_target_dParam[i].atomic_position_derivatives[j];

        nAdps = mAdp2Param[i].size();

        for (j = 0; j<nAdps; j++)
            if(mAdp2Param[i][j] != -1)
                derivatives[ mAdp2Param[i][j] ] = mD_target_dParam[i].adp_derivatives[j];

    }
    double result = calc_wR2_square();
    
    return result;
}

void Target::updateLcs(const Crystal &crystal)
{
    size_t nAtoms = crystal.atoms.size();
    Matrix3d m;
    if(mLcs.empty())
    {
        m.setToIdentity();
        mLcsMatrices.assign(nAtoms, m);
    }
    else
        for(size_t i=0;i<nAtoms;i++)
            mLcs[i].calculate(mLcsMatrices[i],crystal);
    
}

void Target::parameterizeCrystal(
    const std::vector<double> &parameters, 
    Crystal &crystal)
{
    size_t i, j, nAdps, nAtoms = crystal.atoms.size();
    vector<AtomInCrystal> & atoms = crystal.atoms;

    for (i = 0; i< nAtoms; i++)
    {
        if (mOcc2Param[i] != -1)
            atoms[i].occupancy = parameters[mOcc2Param[i]];
        else
            atoms[i].occupancy = mCrystal.atoms[i].occupancy; 


        for (j = 0; j<3; j++)
            if(mXyz2Param[i][j] != -1)
                atoms[i].coordinates[j] = parameters[mXyz2Param[i][j]];
            else 
                atoms[i].coordinates[j] = mCrystal.atoms[i].coordinates[j];

        nAdps = mAdp2Param[i].size();

        for (j = 0; j<nAdps; j++)
            if(mAdp2Param[i][j] != -1)
                atoms[i].adp[j] = parameters[mAdp2Param[i][j]];
            else
                atoms[i].adp[j] = mCrystal.atoms[i].adp[j];
    }

}

/**
set parameters for the weighting scheme
w = 1 / [ sigma2(Fo2) + (aP)2 + bP ]
where P = [ 2Fc2 + Max(Fo2,0) ] / 3
*/
void Target::setWeightParameters(
    double a, 
    double b)
{
    mA = a;
    mB = b;
}

void Target::onlyAdps()
{
    size_t parameterIdx, atomIdx, i, nAdps, nAtoms = mOcc2Param.size();

    parameterIdx = 0;

    for (atomIdx = 0; atomIdx<nAtoms; atomIdx++)
    {
        mOcc2Param[atomIdx] = -1;

        for (i = 0; i<3; i++)
            mXyz2Param[atomIdx][i] = -1;

        nAdps = mAdp2Param[atomIdx].size();

        for (i = 0; i<nAdps; i++)
            mAdp2Param[atomIdx][i] = parameterIdx++;

    }

    mN_Param = parameterIdx;
}

void Target::onlyXyz()
{
    size_t parameterIdx, atomIdx, i, nAdps, nAtoms = mOcc2Param.size();

    parameterIdx = 0;

    for (atomIdx = 0; atomIdx<nAtoms; atomIdx++)
    {
        mOcc2Param[atomIdx] = -1;

        for (i = 0; i<3; i++)
            mXyz2Param[atomIdx][i] = parameterIdx++;

        nAdps = mAdp2Param[atomIdx].size();

        for (i = 0; i<nAdps; i++)
            mAdp2Param[atomIdx][i] = -1;

    }

    mN_Param = parameterIdx;

}

void Target::onlyXyzAndAdps()
{
    size_t parameterIdx, atomIdx, i, nAdps, nAtoms = mOcc2Param.size();

    parameterIdx = 0;

    for (atomIdx = 0; atomIdx<nAtoms; atomIdx++)
    {
        mOcc2Param[atomIdx] = -1;

        for (i = 0; i<3; i++)
            mXyz2Param[atomIdx][i] = parameterIdx++;

        nAdps = mAdp2Param[atomIdx].size();

        for (i = 0; i<nAdps; i++)
            mAdp2Param[atomIdx][i] = parameterIdx++;

    }

    mN_Param = parameterIdx;

}

