#include "Target.h"

#include "discamb/IO/xd_io.h"
#include "discamb/BasicUtilities/Exception.h"
#include "discamb/BasicUtilities/StringUtilities.h"
#include "discamb/BasicUtilities/OnError.h"
#include "discamb/Scattering/HansenCoppensStructureFactorCalculator.h"
#include "discamb/CrystalStructure/StructuralParametersConverter.h"

#include "stdafx.h"
#include "optimization.h"

#include <iostream>
#include <cstdlib> // rand() for randomly distorting structure
#include <ctime>   // for seeding
#include <iomanip>
#include <fstream>



using namespace std;
using namespace discamb;

double random(double min, double max)
{
    double dist = max-min;
    return min + double(rand())/double(RAND_MAX) * dist;
}


double randomOccupancy(double occ, double d_occ)
{
    double result = occ + random(-d_occ, d_occ);
    if(result<0)
        return 0;
    if(result>1)
        return 1;
    return result;

}

double randomADP(double adp, double d_adp)
{
    return adp * ( 100.0 + d_adp * random(-1,1)) / 100.0;
}

// randomly distorts structure

void kick(Crystal &crystal, double d_occ, double d_x, double d_u, int seed=-1)
{
    size_t nAtoms = crystal.atoms.size();
    Vector3d atomicPositionCart, atomicPositionFrac;

    if(seed == -1)
        seed = int(time(NULL));

    srand(seed);

    for( size_t i=0; i<nAtoms; i++ )
    {
        // occupancy
        if(crystal.atoms[i].occupancy!=1)
            crystal.atoms[i].occupancy = randomOccupancy(crystal.atoms[i].occupancy, d_occ);
        
        // adps
        for(size_t j=0 ; j<crystal.atoms[i].adp.size(); j++)
            crystal.atoms[i].adp[j] = randomADP(crystal.atoms[i].adp[j], d_u);


        // xyz

        if(crystal.xyzCoordinateSystem == structural_parameters_convention::fractional)
            crystal.unitCell.fractionalToCartesian(crystal.atoms[i].coordinates, atomicPositionCart);
        else
            atomicPositionCart = crystal.atoms[i].coordinates;

        for(size_t j=0;j<3;j++)
            atomicPositionCart[j] += random( -d_x, d_x );

        if(crystal.xyzCoordinateSystem == structural_parameters_convention::fractional)
        {
            crystal.unitCell.cartesianToFractional(atomicPositionCart, atomicPositionFrac);
            crystal.atoms[i].coordinates = atomicPositionFrac;
        }
        else 
            crystal.atoms[i].coordinates = atomicPositionCart;
    }

}

void function_and_grad(
    const alglib::real_1d_array &x, 
    double &func, 
    alglib::real_1d_array &grad, 
    void *ptr)
{
    Target *t = static_cast<Target*>(ptr);
    size_t n = x.length();
    vector<double> args(n), _grad(n);

    for (size_t i = 0; i<n; i++)
        args[i] = x[i];

    func = t->targetAndDeriveatives(args, _grad);

    for (size_t i = 0; i<n; i++)
        grad[i] = _grad[i];

}

void saveCif(
    const std::string &fName,
    const Crystal &crystal)
{
    ofstream out(fName.c_str());

    out << "data_discambrefine\n"
        << "_cell_angle_alpha  " << setw(8) << setprecision(4) << fixed << crystal.unitCell.alpha() << "\n"
        << "_cell_angle_beta   " << setw(8) << setprecision(4) << fixed << crystal.unitCell.beta() << "\n"
        << "_cell_angle_gamma  " << setw(8) << setprecision(4) << fixed << crystal.unitCell.gamma() << "\n"
        << "_cell_length_a     " << setw(8) << setprecision(4) << fixed << crystal.unitCell.a() << "\n"
        << "_cell_length_b     " << setw(8) << setprecision(4) << fixed << crystal.unitCell.b() << "\n"
        << "_cell_length_c     " << setw(8) << setprecision(4) << fixed << crystal.unitCell.c() << "\n"
        << "loop_\n"
        << "_symmetry_equiv_pos_as_xyz\n";

    string symmOp;

    for (size_t i = 0; i < crystal.spaceGroup.nSymmetryOperations(); i++)
    {
        crystal.spaceGroup.getSpaceGroupOperation(i).get(symmOp);
        out << symmOp << "\n";
    }

    out << "loop_\n"
        << "_atom_site_type_symbol\n"
        << "_atom_site_label\n"
        << "_atom_site_fract_x\n"
        << "_atom_site_fract_y\n"
        << "_atom_site_fract_z\n"
        << "_atom_site_U_iso_or_equiv\n"
        << "_atom_site_adp_type\n"
        << "_atom_site_occupancy\n";
    double u_iso;
    for (size_t i = 0; i < crystal.atoms.size(); i++)
    {
        out << setw(4) << crystal.atoms[i].type
            << setw(8) << crystal.atoms[i].label
            << setw(8) << setprecision(4) << fixed << crystal.atoms[i].coordinates[0]
            << setw(8) << crystal.atoms[i].coordinates[1]
            << setw(8) << crystal.atoms[i].coordinates[2];
        if (crystal.atoms[i].adp.empty())
            u_iso = 0.0;
        else
            u_iso = crystal.atoms[i].adp[0];
        out << setw(8) << u_iso;
        if (crystal.atoms[i].adp.size() == 6)
            out << " Uani ";
        else
            out << " Uiso ";
        out << setw(8) << crystal.atoms[i].occupancy << "\n";
    }

    out << "loop_\n"
        << "_atom_site_aniso_label\n"
        << "_atom_site_aniso_U_11\n"
        << "_atom_site_aniso_U_22\n"
        << "_atom_site_aniso_U_33\n"
        << "_atom_site_aniso_U_12\n"
        << "_atom_site_aniso_U_13\n"
        << "_atom_site_aniso_U_23\n";

    for (size_t i = 0; i < crystal.atoms.size(); i++)
        if (crystal.atoms[i].adp.size() == 6)
        {
            out << setw(8) << crystal.atoms[i].label << " ";
            for (size_t j = 0; j < 6; j++)
                out << setw(8) << crystal.atoms[i].adp[j];
            out << endl;
        }

}


void readIntensities(
    const string &fName,
    vector<Vector3i> &hkl,
    vector<double> &intensities,
    vector<double> &sigma,
    size_t format)
{   
    size_t i_obs_column, sigma_column, min_n_words;
    string line;
    vector<string> words;
    double s, intensity;
    ifstream in(fName.c_str());
    
    // only intensities such that I > nSigmaThreshold * sigma(I)
    // will be used
    double nSigmaThreshold = 2.0; 

    if(format != 5 && format != 6)
        on_error::throwException("invalid specification of intensities data file format", __FILE__ , __LINE__ );

    
    hkl.clear();
    intensities.clear();
    sigma.clear();


    if(format == 5)
    {
        i_obs_column = 3;
        sigma_column = 4;
        min_n_words = 5;
    }
    else // 6
    {
        i_obs_column = 4;
        sigma_column = 5;
        min_n_words = 6;
    }
    
    while(in.good())
    {
        getline(in,line);
        string_utilities::split(line, words);
        if(words.size() >= min_n_words)
        {
            intensity = atof(words[i_obs_column].c_str());
            s = atof(words[sigma_column].c_str());
            if( intensity > nSigmaThreshold * s )
            {
                hkl.push_back(Vector3i(atoi(words[0].c_str()), atoi(words[1].c_str()), atoi(words[2].c_str()) ));
                intensities.push_back(intensity);
                sigma.push_back(s);
            }
        }
    }

}

void optimize(
    Target *target,
    std::vector<double> &params,
    Crystal &crystal)
{

    size_t i, nParams = params.size();

    alglib::real_1d_array x;
    x.setlength(nParams);
    vector<double> derivatives;


    for (i = 0; i<nParams; i++)
        x[i] = params[i];


    double epsg = 0.0000000001;
    double epsf = 0;
    double epsx = 0;
    alglib::ae_int_t maxits = 0;
    alglib::minlbfgsstate state;
    alglib::minlbfgsreport rep;
    cout << "nParams " << nParams << endl;
    alglib::minlbfgscreate(nParams, x, state);
    alglib::minlbfgssetcond(state, epsg, epsf, epsx, maxits);
    alglib::minlbfgssetprecscale(state);


    alglib::minlbfgsoptimize(state, function_and_grad, NULL, target);
    alglib::minlbfgsresults(state, x, rep);

    for (i = 0; i<nParams; i++)
        params[i] = x[i];

    target->parameterizeCrystal(params, crystal);
}

void printStructure(
    ostream &out,
    const Crystal &c)
{
    size_t i, j, nU, nAtoms = c.atoms.size();
    out << setw(8) << "label" << setw(6) << "type" << setw(8) << "occ"
        << setw(8) << "x" << setw(8) << "y" << setw(8) << "z" << endl
        << setw(30) << "U" << endl;
    
    

    for (i = 0; i < nAtoms; i++)
    {
        out << setw(8) << c.atoms[i].label << setw(6) << c.atoms[i].type
            << setw(8) << setprecision(4) << fixed << c.atoms[i].occupancy;
        
        for (j = 0; j < 3; j++)
            out << setw(8) << setprecision(4) << fixed << c.atoms[i].coordinates[j];
        
        nU = c.atoms[i].adp.size();

        if (nU > 0)
        {
            out << endl;
            out << setw(30) << setprecision(4) << fixed << c.atoms[i].adp[0];
            for (j = 1; j < nU; j++)
                out << setw(8) << setprecision(4) << fixed << c.atoms[i].adp[j];
        }
        out << endl;
    }

}

void run(
    const std::string &xdMasterFileName, 
    const std::string &xdParameterFileName,
    const std::string &intensitiesDataFileName,
    int format,
    const string & optionsFile,
    bool noDistort=false)
{
    HC_ModelParameters hcModelParameters;
    Crystal crystal0, crystal;
    vector<Vector3i> hkl_indices;
    vector<XdLocalCoordinateSystem> localCoordinateSystems;
    vector<double> intensities, sigma;
    vector<double> params, derivatives;

    readIntensities(intensitiesDataFileName, hkl_indices, intensities, sigma, format);

    // reads information on model from XD files
    xd_io::read(xdMasterFileName, xdParameterFileName, hcModelParameters, crystal, localCoordinateSystems, true);

    // create structure factor calculator
    HansenCoppensStructureFactorCalculator engine(crystal, hcModelParameters);

    engine.setDerivativesConvention(structural_parameters_convention::fractional, structural_parameters_convention::U_cif);

    Target *target;
    
    if(optionsFile.empty())
        target = new Target(&engine, crystal, localCoordinateSystems);
    else
        target = new Target(&engine, crystal, optionsFile, localCoordinateSystems);

    target->setObservationsAndInit(hkl_indices, intensities, sigma);
    target->getParameters(params, crystal);
    cout<< "wR2 before start " << sqrt(target->calc_wR2_square()) << endl;
    
    if (!noDistort)
    {
        kick(crystal, 0.0, 0.1, 50, 1);

        target->getParameters(params, crystal);
        cout << "wR2 after random distortion " << sqrt(target->targetAndDeriveatives(params, derivatives)) << endl;
    }

    optimize(target, params, crystal);

    target->getParameters(params, crystal);
    cout << "wR2 after refinement " << sqrt(target->targetAndDeriveatives(params, derivatives)) << endl;

    cout << endl << "Structure after refinement" << endl << endl;
    printStructure(cout, crystal);

    saveCif("refined.cif", crystal);

    return;
}


int main(int argc,char *argv[])
{
    
    try
    {
        if(argc != 5 && argc != 6 && argc != 7)
            on_error::throwException("invalid number of arguments", __FILE__,__LINE__);
        
        string optionsFile;
        vector<string> args;
      

        bool no_distort = false;

        for (size_t i = 1; i < argc; i++)
            if (argv[i][0] == '-')
            {
                if (string(argv[i]) == string("-nd"))
                    no_distort = true;
                else
                    on_error::throwException(string("unknown option ") + string(argv[i]), __FILE__, __LINE__);
            }
            else
                args.push_back(argv[i]);

        if (args.size() == 5)
            optionsFile = args[4];
        run(args[0], args[1], args[2], atoi(args[3].c_str()), optionsFile, no_distort);
        
    }
    catch (const Exception & e)
    {
        cout<< e.what();
    }
    catch (const std::exception& e)
    {
        cout<< "standad exception " << e.what() << endl;
    }
    
    return 0;
}
