/*
Parameters of wavefunction for carbon atom taken from Clementi E. and Roetti C.
'Atomic data and nuclear tables' 14, p. 177 (1974) are used to construct spherically averaged 
atomic electron density and calculate the corresponding form factor. The parameters are also
obtained from tables included in the discamb and the corresponding form factor are calculated 
for comparison.
*/

#include "discamb/HC_Model/HC_WfnData.h"
#include "discamb/BasicUtilities/Exception.h"
#include "discamb/Scattering/HansenCoppensStructureFactorCalculator.h"
#include "discamb/Scattering/SlaterTypeOrbitalScattering.h"
//#include "discamb/HC_Model/WfnBankData_ClementiRoetti.h"
#include "discamb/HC_Model/ClementiRoettiData.h"

#include <iostream>
#include <algorithm>
#include <iomanip>

using namespace discamb;

/* The function transforms parameters of linear combination of Slater type functions
   to the form used by discamb. The combination is defined as:
   
   \sum c_i A(n_i,a_i) r^n_i e^(-b_i r)
   
   where:
   
   c_i - coefficients of linear combination
   n_i - integer power of r
   b_i - exponent (in Bohrs)
   A(n_i,a_i) - normalization factor

   The resulting exponents (a_i) are given in 1/Angstrom
   and coefficients (d_i) contain the normalization factors, so the sum can be written as:

   \sum d_i r^n_i e^(-a_i r)


*/

void convertCoefficientAndExponents(
    const std::vector<double> &coefficientsWithNoNormFactor, // c_i
    const std::vector<double> &exponentsBohrs, // b_i
    const std::vector<size_t> &powerR, // n_i
    std::vector<double> &coefficientsIncludingNormFactor, // d_i
    std::vector<double> &exponentsAngstroms //a_i
)
{
    size_t i,n = coefficientsWithNoNormFactor.size();
    double angstromOverBohr = 1.0/0.52917721092;
    coefficientsIncludingNormFactor.resize(n);
    exponentsAngstroms.resize(n);

    for(i=0;i<n;i++)
    {
        exponentsAngstroms[i] = exponentsBohrs[i] * angstromOverBohr;
        coefficientsIncludingNormFactor[i] = coefficientsWithNoNormFactor[i]*
                                             sto_atomic_wfn::stoNormalizationFactor(powerR[i],
                                                                                    exponentsAngstroms[i]);

    }
}

// calculate form factors for given atomic wavefunction

void calculateFormFactor(const HC_WfnBankEntry &atomicWavefunction,
                         const std::vector<double> &hLength, // in Angstroms
                         std::vector<double> &formFactorValues)
{
    // calculate spherically averaged atomic density from wavefunction
    std::vector<double> densityCoefficients, densityExponents;
    std::vector<size_t> desnityPowerR;
    sto_atomic_wfn::atomicStoWfnToSphericalDensity(atomicWavefunction, densityCoefficients, densityExponents, desnityPowerR,
                                                   sto_atomic_wfn::TOTAL_DENSITY);
    // calculate scattering factors
    size_t nPoints = hLength.size();
    formFactorValues.resize(nPoints);

    for (size_t i = 0; i<nPoints; i++)
        formFactorValues[i] = sto_scattering::scatteringSphericalDensity(
                                                      densityCoefficients, densityExponents, 
                                                      desnityPowerR, hLength[i]);

}

void run()
{/*    The following entry describes atomic wave function for carbon atom and 
       it is taken from Clementi E. and Roetti C. Atomic data and nuclear tables 14, p. 177 (1974) 
       
       CARBON 1S(2)2S(2)2P(2), 3P
       T.E.=-0.37688612D+02 P.E.=-0.75376969D+02
       K.E.= 0.37688357D+02 V.T.=-0.20000068D+O1
            S              1S       2S           P             2P
        BASIS / ORB E  -11.32554 -O.70563   BASIS / ORB E   -0.43335
        1S 5.43599       0.93262 -0.20814    2P 0.98073      0.28241
        1S 9.48256       0.06931 -0.01071    2P 1.44361      0.54697
        2S 1.05749       0.00083  0.08099    2P 2.60051      0.23195
        2S 1.52427      -0.00176  0.75045    2P 6.51003      0.01025
        2S 2.68435       0.00559  0.33549
        2S 4.20096       0.00382 -0.14765
*/
    // write the values given above into tables
    std::vector<double> exponentsS_OrbitalsBohrs = { 5.43599, 9.48256, 1.05749, 1.52427, 2.68435, 4.20096 };
    std::vector<double> exponentsP_OrbitalBohrs = { 0.98073, 1.44361, 2.60051, 6.51003 };
    std::vector<double> coefficients1S = { 0.93262, 0.06931, 0.00083, -0.00176, 0.00559, 0.00382 };
    std::vector<double> coefficients2S = { -0.20814, -0.01071, 0.08099, 0.75045, 0.33549, -0.14765 };
    std::vector<double> coefficients2P  = { 0.28241, 0.54697, 0.23195, 0.01025 };
    std::vector<size_t> powerR_S = { 0, 0, 1, 1, 1, 1};
    std::vector<size_t> powerR_P = { 1, 1, 1, 1 };
    
    // --------------------------------------------------------------
    //
    // convert the parameters into form suitable for use with discamb
    //
         
        // define orbitals
    SlaterTypeAtomicOrbitalRdf oneS, twoS, twoP;

    oneS.power_r = powerR_S;
    twoS.power_r = powerR_S;
    twoP.power_r = powerR_P;

        // change exponents units to Angstroms and include normalization factors into 
        // coefficients (see description of the called function for more information)

    convertCoefficientAndExponents(coefficients1S, exponentsS_OrbitalsBohrs, oneS.power_r,
                                   oneS.coefficients, oneS.exponents);
    convertCoefficientAndExponents(coefficients2S, exponentsS_OrbitalsBohrs, twoS.power_r,
                                   twoS.coefficients, twoS.exponents);
    convertCoefficientAndExponents(coefficients2P, exponentsP_OrbitalBohrs, twoP.power_r,
                                   twoP.coefficients, twoP.exponents);

        // define atomic wavefunction
    HC_WfnBankEntry carbonWfn;

    carbonWfn.orbitals = {oneS,twoS,twoP};
    carbonWfn.core_orbitals_indices = {0};
    carbonWfn.valence_orbitals_indices = {1,2};
    carbonWfn.orbital_occupancy = {2,2,2};
    
    //
    //---------------------------------------------------------------------------------------


    // generate array of points (0.01,0.02,..,1.0) where scattering factor will be calculated
    std::vector<double> hLength(101);
    std::generate(hLength.begin(), hLength.end(), []()->double {static int i = -1; i++; return i*0.01; });

    // calculate scattering for the wavefunction defined above

    std::vector<double> formFactors;
    calculateFormFactor(carbonWfn,hLength,formFactors);

    // get wavefunction from tables included in discamb

    HC_WfnBankEntry carbonWfnFromWfnTable;
    ClementiRoettiData clementi_roetti_wfns;

    carbonWfnFromWfnTable = clementi_roetti_wfns.getEntry("C");

    // calculate scattering factors for the above defined wavefunction from the tables 

    std::vector<double> formFactors2;
    calculateFormFactor(carbonWfn, hLength, formFactors);

    // print out the form factors for comparison

    for (size_t i = 0; i<formFactors.size(); i++)
        std::cout<< formFactors[i] << "  " << formFactors2[i] << std::endl;

}

int main()
{


    try
    {
        /*    The following entry describes atomic wave function for carbon atom and
        it is taken from Clementi E. and Roetti C. Atomic data and nuclear tables 14, p. 177 (1974)

        CARBON 1S(2)2S(2)2P(2), 3P
        T.E.=-0.37688612D+02 P.E.=-0.75376969D+02
        K.E.= 0.37688357D+02 V.T.=-0.20000068D+O1
        S              1S       2S           P             2P
        BASIS / ORB E  -11.32554 -O.70563   BASIS / ORB E   -0.43335
        1S 5.43599       0.93262 -0.20814    2P 0.98073      0.28241
        1S 9.48256       0.06931 -0.01071    2P 1.44361      0.54697
        2S 1.05749       0.00083  0.08099    2P 2.60051      0.23195
        2S 1.52427      -0.00176  0.75045    2P 6.51003      0.01025
        2S 2.68435       0.00559  0.33549
        2S 4.20096       0.00382 -0.14765
        */

        // write the values given above into tables
        std::vector<double> exponentsS_OrbitalsBohrs = { 5.43599, 9.48256, 1.05749, 1.52427, 2.68435, 4.20096 };
        std::vector<double> exponentsP_OrbitalBohrs = { 0.98073, 1.44361, 2.60051, 6.51003 };
        std::vector<double> coefficients1S = { 0.93262, 0.06931, 0.00083, -0.00176, 0.00559, 0.00382 };
        std::vector<double> coefficients2S = { -0.20814, -0.01071, 0.08099, 0.75045, 0.33549, -0.14765 };
        std::vector<double> coefficients2P = { 0.28241, 0.54697, 0.23195, 0.01025 };
        std::vector<size_t> powerR_S = { 0, 0, 1, 1, 1, 1 };
        std::vector<size_t> powerR_P = { 1, 1, 1, 1 };

        // --------------------------------------------------------------
        //
        // convert the parameters into form suitable for use with discamb
        //

        // define orbitals
        SlaterTypeAtomicOrbitalRdf oneS, twoS, twoP;

        oneS.power_r = powerR_S;
        twoS.power_r = powerR_S;
        twoP.power_r = powerR_P;

        // change exponents units to Angstroms and include normalization factors into 
        // coefficients (see description of the called function for more information)

        convertCoefficientAndExponents(coefficients1S, exponentsS_OrbitalsBohrs, oneS.power_r,
                                       oneS.coefficients, oneS.exponents);
        convertCoefficientAndExponents(coefficients2S, exponentsS_OrbitalsBohrs, twoS.power_r,
                                       twoS.coefficients, twoS.exponents);
        convertCoefficientAndExponents(coefficients2P, exponentsP_OrbitalBohrs, twoP.power_r,
                                       twoP.coefficients, twoP.exponents);

        // define atomic wavefunction
        HC_WfnBankEntry carbonWfn;

        carbonWfn.orbitals = { oneS,twoS,twoP };
        carbonWfn.core_orbitals_indices = { 0 };
        carbonWfn.valence_orbitals_indices = { 1,2 };
        carbonWfn.orbital_occupancy = { 2,2,2 };

        //
        //---------------------------------------------------------------------------------------


        // generate array of points (0.01,0.02,..,1.0) where scattering factor will be calculated
        std::vector<double> hLength(21);
        std::generate(hLength.begin(), hLength.end(), []()->double {static int i = -1; i++; return i*0.1; });

        // calculate scattering for the wavefunction defined above

        std::vector<double> formFactors;
        calculateFormFactor(carbonWfn, hLength, formFactors);

        // get wavefunction from tables included in discamb

        HC_WfnBankEntry carbonWfnFromWfnTable;
        ClementiRoettiData clementi_roetti_wfns;

        carbonWfnFromWfnTable = clementi_roetti_wfns.getEntry("C");

        // calculate scattering factors for the above defined wavefunction from the tables 

        std::vector<double> formFactors2;
        calculateFormFactor(carbonWfn, hLength, formFactors2);

        // print out the form factors for comparison
        std::cout<< "   |h|      form factor 1  form factor 2" << std::endl;
        for (size_t i = 0; i<formFactors.size(); i++)
            std::cout << std::fixed<< std::setw(8) << hLength[i]
                      << std::setw(14) << std::setprecision(5) << formFactors[i] 
                      << std::setw(14) << std::setprecision(5) << formFactors2[i] << std::endl;
    }
    catch (const Exception & e)
    {
        std::cout << e.what();
    }
    catch (const std::exception& e)
    {
        std::cout << "standad exception " << e.what() << std::endl;
    }

    return 0;
}
