#include "discamb/IO/xd_io.h"
#include "discamb/BasicUtilities/Exception.h"
#include "discamb/Scattering/HansenCoppensStructureFactorCalculator.h"

#include <iostream>

using namespace std;
using namespace discamb;

void readParameters(
    const char *xdMasterFileName, 
    const char *xdParameterFileName, 
    HC_ModelParameters &hcModelParameters)
{
    Crystal crystal;
    vector<XdLocalCoordinateSystem> localCoordinateSystems;
    vector<Matrix3d> localCoordinateSystemMatrices;

    // reads informatoin on model from XD files
    xd_io::read(xdMasterFileName, xdParameterFileName, hcModelParameters, crystal, localCoordinateSystems, true, false);
}

size_t countSphericalAtoms(
    const HC_ModelParameters &hcModelParameters)
{
    size_t result = 0;
    vector<bool> isSphericalType;
    size_t typeIdx, nTypes = hcModelParameters.type_parameters.size();

    isSphericalType.assign(nTypes, true);

    for(typeIdx=0; typeIdx<nTypes; typeIdx++)
        for(auto &p_l: hcModelParameters.type_parameters[typeIdx].p_lm)
            for(auto &p_lm: p_l)
                if(p_lm != 0.0)
                    isSphericalType[typeIdx] = false;

    for(auto typeId: hcModelParameters.atom_to_type_map)
        if(isSphericalType[typeId])
            result++;

    return result;

}

void printParametersInfo(
    HC_ModelParameters &hcModelParameters,
    bool print_all)
{
    vector<size_t> wfnTypeCount, atomTypeCount;
    size_t typeIdx, atomIdx, nWfnTypes, nAtomTypes, nAtoms, nSphericalAtoms;

    nWfnTypes = hcModelParameters.wfn_parameters.size();
    nAtomTypes = hcModelParameters.type_parameters.size();
    nAtoms = hcModelParameters.atom_to_type_map.size();
    nSphericalAtoms = countSphericalAtoms(hcModelParameters);

    cout<< "N Atoms           : " << nAtoms << endl
        << "N wfn types       : " << nWfnTypes << endl
        << "N atom types      : " << nAtomTypes << endl
        << "N spherical atoms : " << nSphericalAtoms << endl;

    if(!print_all)
        return;

    wfnTypeCount.resize(nWfnTypes);
    atomTypeCount.resize(nAtomTypes);

    for(atomIdx=0;atomIdx<nAtoms;atomIdx++)
    {
        wfnTypeCount[hcModelParameters.atom_to_wfn_map[atomIdx]]++;
        atomTypeCount[hcModelParameters.atom_to_type_map[atomIdx]]++;
    }

    cout<< endl << "WFN TYPE COUNT:" << endl << endl;
    for(typeIdx=0;typeIdx<nWfnTypes;typeIdx++)
        cout<< "  "  << hcModelParameters.wfn_parameters[typeIdx].wfnAtomTypeName << "   " << wfnTypeCount[typeIdx] << endl;

    cout << endl << "ATOM TYPE COUNT:" << endl << endl;
    for (typeIdx = 0; typeIdx<nAtomTypes; typeIdx++)
        cout << "  " << typeIdx+1 << "   " << atomTypeCount[typeIdx] << endl;

}

void run(const char *xdMasterFileName, const char *xdParameterFileName, bool print_all)
{
    HC_ModelParameters hcModelParameters;

    readParameters(xdMasterFileName, xdParameterFileName, hcModelParameters);
    printParametersInfo(hcModelParameters,print_all);

}

int main(int argc, char *argv[])
{
    bool print_all;

    try
    {
        if (argc != 3 && argc != 4)
        {
            cout << "expected 2 or 3 arguments: [all] name of xd master file and xd parameter file" << endl;
            exit(0);
        }
        if(argc ==4 )
        {
            print_all = true;
            if(string(argv[1]) != string("all"))
            {
                cout<< "incorrect input, expected 2 or 3 arguments: [all] name of xd master file and xd parameter file" << endl;
                exit(0);
            }
        }
        else
            print_all = false;

        if(print_all)
            run(argv[2], argv[3],print_all);
        else
            run(argv[1], argv[2], print_all);
    }
    catch (const Exception & e)
    {
        cout << e.what();
    }
    catch (const std::exception& e)
    {
        cout << "standad exception " << e.what() << endl;
    }

    return 0;
}
