#ifndef _DISCAMB_BASICUTILITIES_STRINGUTILITIES_H_
#define _DISCAMB_BASICUTILITIES_STRINGUTILITIES_H_

#include <string>
#include <algorithm>
#include <functional>
#include <cctype>
#include <locale>
#include <vector>
#include <sstream>
#include <iomanip>
#include <exception>

namespace discamb {

    /**
    * \addtogroup BasicUtilities BasicUtilities
    * @{

    \brief Utilities for operations on strings
    */


namespace string_utilities
{

enum CharacterType {WHITE_SPACE, DIGIT};

/**
\brief Converts the argument to std::string
*/
template<typename T>
std::string convertToString(T const &);

/**
\brief Converts real number to std::string

\param real real number to be converted

\param precision precision of the number as a string

\param setToFixedOrScientific specifies if the last argument (fixed) will be used to set scientific or fixed format
                                (if false default formatting will be used)
\param fixed sets formating to fixed if true and to scientific if false, the argument is used only if 
               setToFixedOrScientific is set to true
*/

template<typename T>
std::string realToString(T const &real,int precision=6,bool setToFixedOrScientific=false,bool fixed=true);
    
/** 
\brief Converts from std::string to type T
*/
template<typename T>
T convertFromString(std::string const &arg);

/**
\brief Converts from std::string to type T
*/
template<typename T>
void convertFromString(std::string const &,T &);


/**
\brief Trims left side of the string
*/

// $todo should it not be const std::strign &s

static inline std::string &trim_left(std::string &s)
{
    s.erase(s.begin(), std::find_if(s.begin(), s.end(), std::not1(std::ptr_fun<int, int>(std::isspace))));
    return s;
}

/**
\brief Trims right side of the string
*/

// $todo const std::strign &s?

static inline std::string &trim_right(std::string &s)
{
    s.erase(std::find_if(s.rbegin(), s.rend(), std::not1(std::ptr_fun<int, int>(std::isspace))).base(), s.end());
    return s;
}

/**
\brief Trims the string s
*/

// $todo const std::strign &s ?

static inline std::string &trim(std::string &s)
{
    return trim_left(trim_right(s));
}

/**
\brief Splits string into words
\param str string to be splitted
\param vecStr result of the split
\param delimiter char being a delimiter or a function accepting char as an argument an returning value convertible to bool 
                  (e.g. isdigit or isspace)

*/

template < typename T >
void split(const std::string &str,std::vector<std::string> &vecStr,T delimiter);

/**
\brief Splits string into words - specialization for delimiter of type char
\param str string to be splitted
\param vecStr result of the split
\param c char being a delimiter
(e.g. isdigit or isspace)

*/


template <>
inline
void split<char>(const std::string &str,std::vector<std::string> &vecStr,char c)
{
    split(str,vecStr,std::bind1st(std::equal_to<char>(),c));
}

/**
\brief Splits string into words - specialization for delimiter of type enum CharacterType
\param str string to be splitted
\param vecStr result of the split
\param cType character type being a delimiter
(e.g. isdigit or isspace)

*/

template <>
inline
void split<CharacterType>(const std::string &str,std::vector<std::string> &vecStr,CharacterType cType)
{
    if(cType == string_utilities::WHITE_SPACE)
        split<int (int)>(str,vecStr,std::isspace);
}

/**
\brief Splits string into words using space as delimiter
\param str string to be splitted
\param vecStr result of the split

*/

inline
void split(const std::string &str,std::vector<std::string> &vecStr)
{
    split<char>(str,vecStr,' ');
}

/**
\brief Splits string into words
\param str string to be splitted
\param vecStr result of the split
\param delimiters vector of characters treated as delimiters
*/


inline
void split(const std::string &str,std::vector<std::string> &vecStr,const std::vector<char> &delimiters)
{
    std::vector<std::string> words,toSplit;
    size_t n,delimiterIndex,j,nDelimiters=delimiters.size();

    vecStr.assign(1,str);
    
     for( delimiterIndex = 0 ; delimiterIndex < nDelimiters ; delimiterIndex++ )
    {
        vecStr.swap(toSplit);
        n = toSplit.size();
        for(j=0;j<n;j++)
        {
            split<char>(toSplit[j], words, delimiters[delimiterIndex]);
            vecStr.insert(vecStr.end(),words.begin(),words.end());
        }
        toSplit.clear();
    }
}


/**
\brief portable version of getline - recognizes LF , CR+LF , LF+CR and CR as end of line
*/

inline
std::istream &portable_getline(std::istream &is,std::string &line)
{
    line.clear();
    char c;
    while(is.get(c))
    {
        if(c==10)
        {
            is.get(c);
            if(c==13)
                return is;
            is.unget();
            return is;
        }
        if(c==13)
        {
            is.get(c);
            if(c==10)
                return is;
            is.unget();
            return is;
        }
        line += c;
    }

    return is;
}

/** \brief Convers all characters in string to lower case. */

inline
void toLower(const std::string &inputString,std::string &transformedString)
{
    transformedString = inputString;
    std::transform(inputString.begin(),inputString.end(),transformedString.begin(),::tolower);
}

/** \brief Convers all characters in string to upper case. */

inline
void toUpper(const std::string &inputString,std::string &transformedString)
{
    transformedString = inputString;
    std::transform(inputString.begin(),inputString.end(),transformedString.begin(),::toupper);
}

/** \brief Removes all occurences of character from string
\param s string from which the character will be removed
\param c the character to remove
*/

inline std::string removeChar(const std::string &s,char c)
{
    std::string result;
    size_t i,n=s.size();
    for(i=0;i<n;i++)
        if(s[i]!=c)
            result+=s[i];
    return result;
}

/**  \brief Merges strings into one string separating them with delimiter
\param words strings to be merged
\param separator string separator

*/

inline std::string merge(const std::vector<std::string> &words,char separator)
{
    std::string result;
    size_t i,n=words.size();
    for(i=0;i<n;i++)
    {
        if(i!=0)
            result+=separator;
        result+=words[i];
    }
    return result;
}

//############# TEMPLATES IMPLEMENTATION ########################


template<typename T>
std::string convertToString(
T const &anything)
{
    std::stringstream ss;
    std::string result;	
    ss<< anything;
    ss>> result;
    return result;
}

    

template<typename T>
T convertFromString(
std::string const &s)
{
    std::stringstream ss;
    T result;
    ss<< s;
    ss>> result;
    
    return result;   
}



template<typename T>
void convertFromString(
std::string const &s,
T &result)
{
    result=convertFromString<T>(s);
}


template<typename T>
std::string realToString(
T const &number,
int precision,
bool setToFixedOrScientific,
bool fixed)
{
    std::stringstream ss;
    std::string result;

    if(setToFixedOrScientific)
        fixed ? ss.flags(std::ios_base::fixed) : ss.flags(std::ios_base::scientific);

    ss.precision(precision);
    
    ss<< number;
    ss>> result;
    
    return result;   
}


template < typename T >
void split(const std::string &str,std::vector<std::string> &vecStr,T t)
{	
    size_t i,nCharacters=str.size();
    bool previousWasDelimiter = true;
    bool currentIsDelimiter;
        
    vecStr.clear();
        
    for(i=0;i<nCharacters;i++)
    {
        currentIsDelimiter = ( t(int(str[i])) != 0 );
        if( !currentIsDelimiter )
        {
            if(previousWasDelimiter)
                vecStr.resize(vecStr.size()+1);		
            vecStr.back() += str[i]; 
        }
        previousWasDelimiter = 	currentIsDelimiter;
    }
}


}

/**@}*/

}// namespace discamb







#endif /*_DISCAMB_BASICUTILITIES_STRINGUTILITIES_H_*/
