#ifndef _DISCAMB_MATHUTILITIES_REALSPHERICALHARMONICS_H_
#define _DISCAMB_MATHUTILITIES_REALSPHERICALHARMONICS_H_

#include "discamb/MathUtilities/Vector3.h"

#include <cstddef>
#include <vector>

namespace discamb {

    /**
    * \addtogroup MathUtilities MathUtilities
    * @{
    */


namespace real_spherical_harmonics
{

    static const double densityNormalizationMultipliers[5][9] = 
    {
     { 0.0795774715459482},
     { 0.3183098861837907, 0.3183098861837907, 0.3183098861837907},
     { 0.7500000000000036, 0.7500000000000036, 0.2067483357831728, 0.7500000000000036, 0.3750000000000018},
     { 0.4244131815783849, 1.999999999999998, 0.3203330895838574, 0.2448537586029175, 0.3203330895838574, 0.999999999999999, 0.4244131815783849},
     { 1.874999999999999, 1.249999999999994, 0.6611826845788106, 0.4740025188689597, 0.06941752438437611, 0.4740025188689597, 0.3305913422894053, 1.249999999999994, 0.4687499999999998}
    };

    static const double wfnNormalizationMultipliers[5][9] =
    {
     { 0.2820947917738781},
     { 0.4886025119029199, 0.4886025119029199, 0.4886025119029199},
     { 1.092548430592079, 1.092548430592079, 0.31539156525252, 1.092548430592079, 0.5462742152960396},
     { 0.5900435899266435, 2.890611442640554, 0.4570457994644657, 0.3731763325901154, 0.4570457994644657, 1.445305721320277, 0.5900435899266435},
     { 2.503342941796705, 1.77013076977993, 0.9461746957575601, 0.6690465435572892, 0.1057855469152043, 0.6690465435572892, 0.47308734787878, 1.77013076977993, 0.6258357354491761}
    };

    /** Calculaes values of density normalized real spherical harmonics, \p point has to be normalized.*/
    template<int MAX_L>
    void getDensityNormalized(const Vector3d &point,std::vector<std::vector<double> > &values);


    /** Calculaes value of density normalized real spherical harmonics, \p point has to be normalized.*/
    double densityNormalized(const Vector3d &point, int l,int m);

    /** Calculaes values of 'wave-function' normalized ('normal' ones, commonly used in math in physics) 
    real spherical harmonics, \p point has to be normalized.*/
    template<int MAX_L>
    void getWfnNormalized(const Vector3d &point,std::vector<std::vector<double> > &values);

    /** Calculaes value of 'wave-function' normalized ('normal' ones, commonly used in math in physics)
    real spherical harmonics, \p point has to be normalized.*/

    double wfnNormalized(const Vector3d &normalizedVector3D, int l, int m);
    
    double polynomial(const Vector3d &noralizedVector3D, int l, int m);

    /** spherical harmonic y_lm can be written as:
        y_lm = r^(-l) * a_lm * w_lm(x,y,z)
        where a_lm is a const. multiplier and w_lm(x,y,z) is a polynomial with integer coefficients

        normalizedPoint - it is expected to be normalized
        values - are expected to be allocated
    */
    template<int MAX_L>
    void getPolynomials(const Vector3d &normalizedPoint, std::vector<std::vector<double> > &values);
    void getMultipliers_DensityNormalization(size_t maxL, std::vector<std::vector<double> > &values);
    void getMultipliers_WfnNormalization(size_t maxL, std::vector<std::vector<double> > &values);
    void getDensityToWfnMultipliers(size_t maxL,std::vector<std::vector<double> > &values);
    //void getWfnToDensityMultipliers(size_t maxL,std::vector<std::vector<double> > &values);


    // TEMPLATE IMPLEMENTATION


    template<int MAX_L>
    inline void getDensityNormalized(
        const Vector3d &point, 
        std::vector<std::vector<double> > &values)
    {
        const int nL = MAX_L + 1;
        //static const vector<vector<double> > wfnNormalizationFactors = getWfnNormalizationFactors();

        getPolynomials<MAX_L>(point, values);

        for (int l = 0; l<nL; l++)
            for (int m = -l; m <= l; m++)
                values[l][l + m] *= densityNormalizationMultipliers[l][l + m];
    }

    template<int MAX_L>
    inline void getWfnNormalized(
        const Vector3d &point, 
        std::vector<std::vector<double> > &values)
    {
        const int nL = MAX_L + 1;
        //static const vector<vector<double> > wfnNormalizationFactors = getWfnNormalizationFactors();

        getPolynomials<MAX_L>(point, values);

        for (int l = 0; l<nL; l++)
            for (int m = -l; m <= l; m++)
                values[l][l + m] *= wfnNormalizationMultipliers[l][l + m];

    }

    template<int maxL>
    inline void getPolynomials(
        const Vector3d &normalizedPoint, 
        std::vector<std::vector<double> > &values)
    {
        if (maxL<0)
            return;

        values[0][0] = 1.0;

        if (maxL > 0)
        {

            const double x = normalizedPoint[0];
            const double y = normalizedPoint[1];
            const double z = normalizedPoint[2];

            values[1][0] = y;
            values[1][1] = z;
            values[1][2] = x;


            if (maxL > 1)
            {

                double xx = x*x;
                double yy = y*y;
                double zz = z*z;
                double xy = x*y;
                double xz = x*z;
                double yz = y*z;
                double xx_p_yy = xx + yy;
                double xx_m_yy = xx - yy;
                double two_zz = 2 * zz;

                values[2][0] = xy;
                values[2][1] = yz;
                values[2][2] = two_zz - xx_p_yy;
                values[2][3] = xz;
                values[2][4] = xx_m_yy;


                if (maxL > 2)
                {
                    double _3xx_m_yy = 3 * xx - yy;
                    double xx_m_3yy = xx - 3 * yy;
                    double four_zz = 4 * zz;

                    values[3][0] = _3xx_m_yy * y;
                    values[3][1] = xy*z;
                    values[3][2] = y * (four_zz - xx_p_yy);
                    values[3][3] = z*(two_zz - 3 * xx_p_yy);
                    values[3][4] = x * (four_zz - xx_p_yy);
                    values[3][5] = (xx_m_yy)*z;
                    values[3][6] = xx_m_3yy * x;


                    if (maxL > 3)
                    {
                        double _7zz = 7 * zz;

                        values[4][0] = xy * xx_m_yy;
                        values[4][1] = _3xx_m_yy * yz;
                        values[4][2] = xy * (_7zz - 1);
                        values[4][3] = yz * (_7zz - 3);
                        values[4][4] = (_7zz - two_zz)*(_7zz - 6) + 3;
                        values[4][5] = xz * (_7zz - 3);
                        values[4][6] = (xx_m_yy)*(_7zz - 1);
                        values[4][7] = xx_m_3yy*xz;
                        values[4][8] = xx * xx_m_3yy - yy * _3xx_m_yy;

                    } // maxL>3
                } // maxL>2
            } // maxL>1
        } // maxL>0

    } //getPolynomials

}
/** @} */
} // namespace discamb

#endif /*_DISCAMB_MATHUTILITIES_REALSPHERICALHARMONICS_H_*/
