#include "discamb/HC_Model/XdLocalCoordinateSystem.h"

#include "discamb/BasicUtilities/OnError.h"

#include <map>
#include <cmath>

using namespace std;

namespace discamb {

XdLocalCoordinateSystem::XdLocalCoordinateSystem()
{
    mIsCartesian = true;
}


XdLocalCoordinateSystem::~XdLocalCoordinateSystem()
{
}
   


void XdLocalCoordinateSystem::set(
const std::string &centralAtom,
const std::string &direction_1_atom,
const std::string &direction_2_atom_1,
const std::string &direction_2_atom_2,
const std::string &vector_1,
const std::string &vector_2,
bool rightHanded,
const Crystal &crystal,
bool dummyAtomIndexingFrom0,
const std::vector<Vector3d> &dummyAtomFarctionalPositions)
{
    map<string,int> directionIndexMap;
    directionIndexMap["X"]=1;
    directionIndexMap["Y"]=2;
    directionIndexMap["Z"]=3;
    directionIndexMap["-X"]=-1;
    directionIndexMap["-Y"]=-2;
    directionIndexMap["-Z"]=-3;

    mIsCartesian = false;
    mCoordinate1VectorMultiplier = directionIndexMap[vector_1]/abs(directionIndexMap[vector_1]);
    mCoordinate1 = abs(directionIndexMap[vector_1]);
    mDirection12Multuiplier = directionIndexMap[vector_2] / abs(directionIndexMap[vector_2]);
    mCoordinate2 = abs(directionIndexMap[vector_2]);
    mCoordinate3 = 6 - mCoordinate1 - mCoordinate2;

    vector<size_t> reference_atom;
    vector<SpaceGroupOperation> aux_sym_ops;

    getRefAtomIndex(centralAtom,reference_atom,aux_sym_ops,crystal);
    mRefAtom = reference_atom[0];

    processAtomLabel(direction_1_atom,crystal,dummyAtomIndexingFrom0,dummyAtomFarctionalPositions,
                     mIsDirection1AtomDummy,mDummy_1_Position,mDirection1Atoms,mDirection1AtomsSymmOps);

    processAtomLabel(direction_2_atom_1,crystal,dummyAtomIndexingFrom0,dummyAtomFarctionalPositions,
                     mIsDirection2Atom1Dummy,mDummy_2_1_Position,mDirection2Atoms1,mDirection2Atoms1SymmOps);

    processAtomLabel(direction_2_atom_2,crystal,dummyAtomIndexingFrom0,dummyAtomFarctionalPositions,
                     mIsDirection2Atom2Dummy,mDummy_2_2_Position,mDirection2Atoms2,mDirection2Atoms2SymmOps);

        
    // 1 for R, -1 for L
    rightHanded ? mHandedeness = 1 : mHandedeness = -1;  

    mAxesOrderRelatedHandedenessCorrection = 1 - 2*( (2+mCoordinate2-mCoordinate1) %3);
}

void XdLocalCoordinateSystem::processAtomLabel(
const std::string &atomLabel,
const Crystal &crystal,
bool dummyAtomIndexingFrom0,
const std::vector<Vector3d> &dummyAtomPositionsFract,
bool &isDummyAtom,
Vector3d &dummyAtomPositionCartesian,
std::vector<size_t> &atomIndices,
std::vector<SpaceGroupOperation> &symmetryOperations)
{

    isDummyAtom =  isDummyAtomLabel(atomLabel);

    if(isDummyAtom)
        crystal.unitCell.fractionalToCartesian(dummyAtomPositionsFract[dummyAtomIndex(atomLabel,dummyAtomIndexingFrom0)],
                                               dummyAtomPositionCartesian);
    else
        getRefAtomIndex(atomLabel,atomIndices,symmetryOperations,crystal);
}


void XdLocalCoordinateSystem::getRefAtomIndex(
const std::string &label,
std::vector<size_t> &atomIndices,
std::vector<SpaceGroupOperation> &symmetryOperations,
const Crystal &c)
{
    size_t i,j,nAtomsInDefinition,nAtomsInCrystal=c.atoms.size();
    vector<string> atomsAsStrings;
    int unassignedAtom;
    bool atomAssigned;
    string_utilities::split(label,atomsAsStrings,';');

    nAtomsInDefinition = atomsAsStrings.size();

    atomIndices.resize(nAtomsInDefinition);
    symmetryOperations.assign(nAtomsInDefinition,SpaceGroupOperation());

    unassignedAtom = -1;
    for(i=0;i<nAtomsInDefinition;i++)
    {
        atomAssigned = false;
        for(j=0;j<nAtomsInCrystal;j++)
            if (c.atoms[j].label == atomsAsStrings[i])
            {
                atomIndices[i] = j;
                atomAssigned = true;
            }

        if(!atomAssigned)
            unassignedAtom = i;
    }

    if( unassignedAtom > -1 )
    {
        string errorMessage = string("unknown atom label '");
        errorMessage += atomsAsStrings[unassignedAtom];
        errorMessage += string("' in definition of local coordinate system for aspherical atom in Hansen-Coppens model.");
        on_error::throwException(errorMessage, __FILE__, __LINE__);
    }
    
}

bool XdLocalCoordinateSystem::isDummyAtomLabel(
const std::string &label)
{
    if(label.size()<4)
        return false;
    return (string("DUM")==label.substr(0,3));
}

size_t XdLocalCoordinateSystem::dummyAtomIndex(
const std::string &label,
bool dummyAtomIndexingFrom0)
{
    int index = atoi(label.substr(3).c_str());
    if(!dummyAtomIndexingFrom0)
        --index;
    return index;
}



void XdLocalCoordinateSystem::calculate(
Vector3d &x,
Vector3d &y,
Vector3d &z,
const Crystal &crystal) 
const
{
    Vector3d point_1_position,point_2_position,point_3_position,reference_atom_position;
    Vector3d direction[3],direction_12;
    Vector3d *axes[]={&x,&y,&z};
    
    if (mIsCartesian)
    {
        x = Vector3d(1, 0, 0);
        y = Vector3d(0, 1, 0);
        z = Vector3d(0, 0, 1);
        return;
    }

    if(mIsDirection1AtomDummy)
        point_1_position = mDummy_1_Position;
    else
        point_1_position = averageAtomSetPosition(crystal,mDirection1Atoms,mDirection1AtomsSymmOps);
        

    if(mIsDirection2Atom1Dummy)
        point_2_position = mDummy_2_1_Position;
    else
        point_2_position = averageAtomSetPosition(crystal,mDirection2Atoms1,mDirection2Atoms1SymmOps);

    if(mIsDirection2Atom2Dummy)
        point_3_position = mDummy_2_2_Position;
    else
        point_3_position = averageAtomSetPosition(crystal,mDirection2Atoms2,mDirection2Atoms2SymmOps);

    if(crystal.xyzCoordinateSystem==structural_parameters_convention::fractional)
        crystal.unitCell.fractionalToCartesian(crystal.atoms[mRefAtom].coordinates, reference_atom_position);
    else
        reference_atom_position = crystal.atoms[mRefAtom].coordinates;


    direction[0] = mCoordinate1VectorMultiplier * (point_1_position - reference_atom_position);
    direction[0] /= sqrt(direction[0]*direction[0]);
     
    direction_12 = mDirection12Multuiplier * (point_3_position - point_2_position);
     
    direction[2] = mAxesOrderRelatedHandedenessCorrection * mHandedeness * cross_product(direction[0],direction_12);
    direction[2] /= sqrt(direction[2]*direction[2]);
     
    direction[1] = mAxesOrderRelatedHandedenessCorrection * mHandedeness * cross_product(direction[2], direction[0]);

    *axes[mCoordinate1-1] = direction[0];
    *axes[mCoordinate2-1] = direction[1];
    *axes[mCoordinate3-1] = direction[2];   
}

Vector3d XdLocalCoordinateSystem::averageAtomSetPosition(
const Crystal &crystal,
const std::vector<size_t> &atoms,
const std::vector<SpaceGroupOperation> &symmetryOperations)
{
    Vector3d atom_fractional, position_fractional,average_position_fractional,result;
    size_t atom_index,nAtoms = atoms.size();

    for(atom_index=0;atom_index<nAtoms;atom_index++)
    {
        // get position in fractional coordinates
        if(crystal.xyzCoordinateSystem == structural_parameters_convention::fractional)
            atom_fractional = crystal.atoms[atoms[atom_index]].coordinates;
        else
            crystal.unitCell.cartesianToFractional(crystal.atoms[atoms[atom_index]].coordinates, atom_fractional);
        symmetryOperations[atom_index].apply(atom_fractional, position_fractional);
        average_position_fractional += position_fractional;
    }

    crystal.unitCell.fractionalToCartesian(average_position_fractional,result);

    return result;
}

} // namespace discamb

