#include "GradientDataCollector.h"
#include "discamb/CrystalStructure/StructuralParametersConverter.h"

using namespace std;
using namespace discamb;

namespace{

    void convertAdps(
        const std::vector<std::vector<std::complex<double> > > &adpsIn,
        std::vector<std::vector<std::complex<double> > > &adpsOut,
        void (StructuralParametersConverter::*convertingFunction)(
            const vector<complex<double> > &derivativesIn,
            vector<complex<double> > &derivativesOut) const,
        StructuralParametersConverter &converter)
    {
        vector<complex<double> > vIn(6), vOut(6);
        size_t hklIdx, nHkl, i;

        nHkl = adpsIn[0].size();

        for (hklIdx = 0; hklIdx<nHkl; hklIdx++)
        {
            for (i = 0; i<6; i++)
                vIn[i] = adpsIn[i][hklIdx];

            (converter.*convertingFunction)(vIn, vOut);

            for (i = 0; i<6; i++)
                adpsOut[i][hklIdx] = vOut[i];
        }

    }


}

void GradientDataSM::allocate(
    size_t nHkl,
    const std::vector<size_t> &nAdpsPerAtom)
{
    size_t atomIdx, nAtoms = nAdpsPerAtom.size();

    clear();
    atomLabel.resize(nAtoms);
    dOccupancy.resize(nAtoms, vector<complex<double> >(nHkl));
    dPosition.resize(nAtoms, vector<vector<complex<double> > >(3, vector<complex<double> >(nHkl)));

    dU_iso.resize(nAtoms);
    dU_aniso.resize(nAtoms);

    for (atomIdx = 0; atomIdx<nAtoms; atomIdx++)
        if (nAdpsPerAtom[atomIdx] == 1)
            dU_iso[atomIdx].resize(nHkl);
        else if (nAdpsPerAtom[atomIdx] == 6)
            dU_aniso[atomIdx].resize(6, vector<complex<double> >(nHkl));

}

void GradientDataSM::clear()
{
    atomLabel.clear();
    dOccupancy.clear();
    dU_iso.clear();
    dPosition.clear();
    dU_aniso.clear();
}

//---------------------------------------------------

void GradientDataSM_AllConventions::clear()
{
    hasXyzCart = hasXyzFract = hasUcart = hasUstar = hasUcif = false;

    atomLabel.clear();
    occupancy.clear();
    u_iso.clear();
    position_fractional.clear();
    position_cartesian.clear();
    u_cartesian.clear();
    u_star.clear();
    u_cif.clear();
}

void GradientDataSM_AllConventions::set(
    const GradientDataSM &data,
    const UnitCell &unitCell)
{
    StructuralParametersConverter derivativesConverter;
    derivativesConverter.set(unitCell);
    size_t atomIdx, nAtoms, nHkl, i;
    Vector3<complex<double> > vec3cIn, vec3cOut;
    vector<complex<double> > vec6cIn(6), vec6cOut(6);


    nAtoms = data.atomLabel.size();

    if (nAtoms == 0)
        return;

    nHkl = data.dOccupancy[0].size();
    atomLabel = data.atomLabel;
    occupancy = data.dOccupancy;
    u_iso = data.dU_iso;
    position_cartesian = data.dPosition;
    u_cartesian = data.dU_aniso;

    // used only for allocation purposes

    position_fractional = data.dPosition;
    u_cif = data.dU_aniso;
    u_star = data.dU_aniso;

    //

    hasUcart = hasXyzCart = true;
    hasXyzFract = hasUstar = hasUcif = false;

    /*
    the code below assumes that there can be other situation than the above
    hasUcart = hasXyzCart = true;
    hasXyzFract = hasUstar = hasUcif = false;
    */

    for (atomIdx = 0; atomIdx<nAtoms; atomIdx++)
    {
        if (!hasXyzCart && hasXyzFract)
        {
            for (size_t hklIdx = 0; hklIdx<nHkl; hklIdx++)
            {
                for (i = 0; i<3; i++)
                    vec3cIn[i] = position_fractional[atomIdx][i][hklIdx];
                derivativesConverter.derivativesXyzFractionalToCartesian(vec3cIn, vec3cOut);
            }
        }
        if (hasXyzCart && !hasXyzFract)
        {
            for (size_t hklIdx = 0; hklIdx<nHkl; hklIdx++)
            {
                for (i = 0; i<3; i++)
                    vec3cIn[i] = position_cartesian[atomIdx][i][hklIdx];
                derivativesConverter.derivativesXyzCartesianToFractional(vec3cIn, vec3cOut);

                for (i = 0; i<3; i++)
                    position_fractional[atomIdx][i][hklIdx] = vec3cOut[i];
            }
        }

        if (hasUcart && u_cartesian[atomIdx].size() == 6)
        {

            if (!hasUstar)
                convertAdps(u_cartesian[atomIdx], u_star[atomIdx],
                            &StructuralParametersConverter::derivativesU_CartesianToU_Star, derivativesConverter);

            if (!hasUcif)
                convertAdps(u_cartesian[atomIdx], u_cif[atomIdx],
                            &StructuralParametersConverter::derivativesU_CartesianToU_Cif, derivativesConverter);
        }

        if (hasUstar && u_star[atomIdx].size() == 6)
        {

            if (!hasUcart)
                convertAdps(u_star[atomIdx], u_cartesian[atomIdx],
                            &StructuralParametersConverter::derivativesU_StarToU_Cartesian, derivativesConverter);

            if (!hasUcif)
                convertAdps(u_star[atomIdx], u_cif[atomIdx],
                            &StructuralParametersConverter::derivativesU_StarToU_Cif, derivativesConverter);
        }

        if (hasUcif && u_cif[atomIdx].size() == 6)
        {

            if (!hasUstar)
                convertAdps(u_cif[atomIdx], u_star[atomIdx],
                            &StructuralParametersConverter::derivativesU_CifToU_Star, derivativesConverter);

            if (!hasUcart)
                convertAdps(u_cif[atomIdx], u_cartesian[atomIdx],
                            &StructuralParametersConverter::derivativesU_CifToU_Cartesian, derivativesConverter);
        }


    }

}




//---------------------------------------------------

GradientDataCollector::GradientDataCollector(
    size_t nHkl,
    const std::vector<size_t> &nAdpsPerAtom)
{
    mGradientData.allocate(nHkl, nAdpsPerAtom);
    mStructureFactors.assign(nHkl, 0.0);
}

GradientDataCollector::~GradientDataCollector()
{
}

void GradientDataCollector::onPerHklCalculation(
    size_t hklIdx,
    std::complex<double> &structureFactor,
    SfDerivativesAtHkl &derivatives)
{
    size_t atomIdx, nAtoms = mGradientData.dOccupancy.size();

    mStructureFactors[hklIdx] = structureFactor;

    for (atomIdx = 0; atomIdx<nAtoms; atomIdx++)
    {
        mGradientData.dOccupancy[atomIdx][hklIdx] = derivatives.occupancyDerivatives[atomIdx];

        for (size_t i = 0; i<3; i++)
            mGradientData.dPosition[atomIdx][i][hklIdx] = derivatives.atomicPostionDerivatives[atomIdx][i];

        if (derivatives.adpDerivatives[atomIdx].size() == 1)
            mGradientData.dU_iso[atomIdx][hklIdx] = derivatives.adpDerivatives[atomIdx][0];

        if (derivatives.adpDerivatives[atomIdx].size() == 6)
        {
            for (size_t i = 0; i<6; i++)
                mGradientData.dU_aniso[atomIdx][i][hklIdx] = derivatives.adpDerivatives[atomIdx][i];
        }
    }

}


void GradientDataCollector::getGradientData(
    GradientDataSM &gradientData)
{
    gradientData = mGradientData;
}

void GradientDataCollector::getStructureFactors(
    std::vector<std::complex<double> > &scatteringFactors)
{
    scatteringFactors = mStructureFactors;
}
