#include "Test_CpuGpu.h"
#include "test_utilities.h"
#include "data_directory.h"

#include "discamb/MathUtilities/Vector3.h"

#include "discamb/BasicUtilities/OnError.h"
#include "discamb/BasicUtilities/StringUtilities.h"
#include "discamb/CrystalStructure/ReciprocalLatticeUnitCell.h"
#include "discamb/Scattering/HansenCoppensStructureFactorCalculator.h"


#include "discamb/IO/xd_io.h"
#include "discamb/IO/hkl_io.h"


#include <fstream>
#include <iomanip>

#ifdef _OPENMP
#include <omp.h>
#endif

using namespace std;
using namespace discamb;





TestCpuGpu::TestCpuGpu()
{
    mN_Cores=0;
}

TestCpuGpu::~TestCpuGpu()
{
}



bool TestCpuGpu::run_test(
    std::string &message, 
    const std::vector<std::string> &arguments)
{
    bool nonDefaultDir = !arguments.empty();
    string s, inputPath, inputDir, dataDir, dTarge_dF_File;
    vector<string> hklFiles, xdMasFiles, xdInpFiles, failedSystems, passedSystems;
    vector<complex<double> > dTarget_dF, sizeAdjusted_dTarget_dF;
    size_t nDataSets;

    if(nonDefaultDir)
        inputDir = arguments[0] + string("/");
    else
        inputDir = correcness_test_1_data_directory + string("/");

    inputPath = inputDir + string("/test_cpu_gpu.inp");
    dataDir = correcness_test_1_data_directory;

    ifstream in(inputPath.c_str());


    if(!in.good())
        on_error::throwException("can not read input file for test comparing CPU calculations with GPU ones", __FILE__, __LINE__);
    
    // n CPU cores: 8
    in>> s >> s >> s >> mN_Cores;
    // dTarget/dF file : dTarget_dF
    in>> s >> s >> dTarge_dF_File;
    //n datasets: 1
    in>> s >> s >> nDataSets;

    if(nonDefaultDir)
        test_utilities::read_dTarget_dSF(inputDir + string("/") + dTarge_dF_File, dTarget_dF);
    else
        test_utilities::read_dTarget_dSF(dataDir + string("/") + dTarge_dF_File, dTarget_dF);

    hklFiles.resize(nDataSets);
    xdMasFiles.resize(nDataSets); 
    xdInpFiles.resize(nDataSets);

    bool passed;

    for(size_t i=0; i<nDataSets; i++)
    {
        in>> xdMasFiles[i] >> xdInpFiles[i] >> hklFiles[i];

        if(nonDefaultDir)
            passed = calculateAndCompare(inputDir + string("/") + xdMasFiles[i],
                             inputDir + string("/") + xdInpFiles[i],
                             inputDir + string("/") + hklFiles[i],
                             dTarget_dF);
        else
            passed = calculateAndCompare(dataDir + string("/structure_data/xd/") + xdMasFiles[i],
                             dataDir + string("/structure_data/xd/") + xdInpFiles[i], 
                             dataDir + string("/hkl_indices/") + hklFiles[i],
                             dTarget_dF);

        if(passed)
            passedSystems.push_back(xdMasFiles[i]);
        else
            failedSystems.push_back(xdMasFiles[i]);
    }


    if (mMessageLevel>0)
    {
    
		if (!failedSystems.empty())
		{
			cout << "systems for which the cpu gpu test passed:" << endl;
			for (size_t i = 0; i < passedSystems.size(); i++)
				cout << passedSystems[i] << endl;

			cout << "systems for which the cpu gpu test failed:" << endl;
			for (size_t i = 0; i < failedSystems.size(); i++)
				cout << failedSystems[i] << endl;
		}

        if(failedSystems.empty())
            cout<< "CPU vs GPU test passed" << endl;
        else
            cout << "CPU vs GPU test failed" << endl;

    }

    return failedSystems.empty();

}

bool TestCpuGpu::calculateAndCompare(
    std::string xdMas, 
    std::string xdInp, 
    std::string hklFile,
    const std::vector<std::complex<double> > &_dTarget_dF)
{
    vector<complex<double> > f_CPU, f_GPU;
    Crystal crystal;
    HC_ModelParameters hc_parameters;
    std::vector<TargetFunctionAtomicParamDerivatives> dTarg_dParam_CPU, dTarg_dParam_GPU;
    vector<XdLocalCoordinateSystem> lcs;
    xd_io::read(xdMas, xdInp, hc_parameters, crystal, lcs, true, false);
    vector<bool> countAtomContribution(crystal.atoms.size(), true);
    vector<Vector3i> hkl;
    hkl_io::readCompactHklFile(hklFile, hkl);
    vector<complex<double> > dTarget_dF;

    dTarget_dF.insert(dTarget_dF.end(), _dTarget_dF.begin(), _dTarget_dF.begin() + hkl.size());

    vector<Matrix3d> lcsMatrices;

    lcsMatrices.resize(lcs.size());
    for (size_t i = 0; i< lcs.size(); i++)
        lcs[i].calculate(lcsMatrices[i], crystal);

    HansenCoppensStructureFactorCalculator hcCalculator(crystal, hc_parameters);
    hcCalculator.setDerivativesConvention(structural_parameters_convention::cartesian, structural_parameters_convention::U_cart);
    hcCalculator.setN_Threads(mN_Cores);

    hcCalculator.calculateStructureFactorsAndDerivatives(crystal.atoms, lcsMatrices, hkl, f_CPU,
                                                         dTarg_dParam_CPU, dTarget_dF, countAtomContribution);
    hcCalculator.useGPU();
    //hcCalculator.useCPU();
    hcCalculator.calculateStructureFactorsAndDerivatives(crystal.atoms, lcsMatrices, hkl, f_GPU,
                                                         dTarg_dParam_GPU, dTarget_dF, countAtomContribution);

    return compare(f_CPU, f_GPU, dTarg_dParam_CPU, dTarg_dParam_GPU);
    
}

bool TestCpuGpu::compareDerivatives(
    const std::string &label, 
    const std::vector<double> &d1, 
    const std::vector<double> &d2)
{
    double rfac = test_utilities::calculateR_Factor(d1,d2,false);
    if(mMessageLevel>1)
        cout<< "rfac for derivative w.r.t. " << label << " " << rfac << endl;
    return rfac<1e-6;
}

bool TestCpuGpu::compare(
    const std::vector<std::complex<double> > &f1,
    const std::vector<std::complex<double> > &f2,
    const std::vector<discamb::TargetFunctionAtomicParamDerivatives> &d1,
    const std::vector<discamb::TargetFunctionAtomicParamDerivatives> &d2)
{
    bool result = true;
    double rFac;
    size_t i, j, nAtoms;
    vector<double> v1,v2;

    // structure factors

    rFac = test_utilities::calculateR_Factor(f1, f2, false);
    if(rFac > 1e-6)
        result = false;
    if(mMessageLevel>1)
        cout<< "r fac " << rFac << endl;

    // derivatives

    nAtoms = d1.size();
    v1.resize(nAtoms);
    v2.resize(nAtoms);

    // occupancy
    
    for(i=0;i<nAtoms;i++)
    {
        v1[i] = d1[i].occupancy_derivatives;
        v2[i] = d2[i].occupancy_derivatives;
    }

    result = result && compareDerivatives("occupancy",v1,v2);

    // xyz
    string xyz[] = {"x","y","z"};

    for(i=0;i<3;i++)
    {
        for(j=0;j<nAtoms;j++)
        {
            v1[j] = d1[j].atomic_position_derivatives[i];
            v2[j] = d2[j].atomic_position_derivatives[i];
        }

        result = result && compareDerivatives(xyz[i], v1, v2);
    }

    // U_ij
    string u_ij[] = { "U_11", "U_22", "U_33", "U_12", "U_13", "U_23"};

    for (i = 0; i<6; i++)
    {
        v1.clear();
        v2.clear();

        for (j = 0; j<nAtoms; j++)
            if(d1[j].adp_derivatives.size()==6)
            {
                v1.push_back(d1[j].adp_derivatives[i]);
                v2.push_back(d2[j].adp_derivatives[i]);
            }

        result = result && compareDerivatives(u_ij[i], v1, v2);
    }

    // U_iso
    v1.clear();
    v2.clear();

    for (j = 0; j<nAtoms; j++)
        if (d1[j].adp_derivatives.size() == 1)
        {
            v1.push_back(d1[j].adp_derivatives[0]);
            v2.push_back(d2[j].adp_derivatives[0]);
        }

    result = result && compareDerivatives("U iso", v1, v2);
    
    return result;
}

