#include "Test_GradientLarge.h"
#include "test_utilities.h"
#include "data_directory.h"

#include "discamb/MathUtilities/Vector3.h"

#include "discamb/BasicUtilities/OnError.h"
#include "discamb/BasicUtilities/StringUtilities.h"
#include "discamb/CrystalStructure/ReciprocalLatticeUnitCell.h"


#include "discamb/CrystalStructure/StructuralParametersConverter.h"

#include "discamb/IO/NativeIAM_Reader.h"
#include "discamb/IO/hkl_io.h"

#include <iostream>
#include <fstream>

using namespace std;
using namespace discamb;



namespace{

    struct GradientR_Factor
    {
        double occupancy;
        double position_fractional[3];
        double position_cartesian[3];
        double u_fractional[3];
        double u_cartesian[3];
        double u_star[3];
        double u_iso;
    };

}


TestGradientLarge::TestGradientLarge()
{
    mWrite = false;
}

TestGradientLarge::~TestGradientLarge()
{
}


bool TestGradientLarge::run_test(
    std::string &message)
{
    return run_test(message, false);
}

bool TestGradientLarge::run_test(
    std::string & message,
    const std::vector<std::string>& _arguments)
{
    vector<string> arguments;// = _arguments;
    bool isOption;

    for(size_t i=0;i<_arguments.size(); i++)
    {
        isOption = false;
        if(_arguments[i].size()>2)
        {
            if(_arguments[i].substr(0,2)==string("--"))
                isOption=true;
        }
        
        if(isOption)
        {
            if(_arguments[i]=="--write")
                mWrite = true;
        }
        else
            arguments.push_back(_arguments[i]);
        
    }


    if (arguments.size() == 0)
        return run_test(message, false);
    
    return run_test(message, true, arguments[0]);
}

std::string TestGradientLarge::makePrefix(
    bool oneDir, 
    const std::string oneDirPath, 
    const std::string defaultPath)
{
    if(oneDir)
    {
        if(oneDirPath.empty())
            return string();
        else 
            return oneDirPath + string("/");
    }
    else
        return defaultPath + string("/");
}

bool TestGradientLarge::run_test(
    std::string &message, 
    bool customDir, 
    const std::string &customDirPath)
{
    string defaultMainDir = correcness_test_1_data_directory;
    message.clear();
    string inputFilePath;

    inputFilePath = makePrefix(customDir, customDirPath, defaultMainDir) + inputFileName();
    
    string s, dTarget_dF_FileName, line;
    string prefixHkl = makePrefix(customDir, customDirPath, defaultMainDir + string("/hkl_indices"));

    
    /* r_factors[i].first - name of i-th data set, 
       r_factors[i].second - R - factor for i-th data set */
    vector<pair<string, GradientData> > r_factors;
    vector<string> setsWhichPassedTest,setsWhichFailedTest;

    vector<vector<string> > structureFiles;
    vector<string> hklFiles, gradientFiles;
        

    /* testValidity[i].parameter_value is set to 1 
       if the dataset for the parameter contains numbers effectively different than 0
       (i.e. fabs(data[i]) > threshold, threshold defined in the hasNonzeroElement method)
       otherwise it is set to 0 (and valid R-factor test cannot be performed */

    vector<GradientData> testValidity;
    vector<string> words;
    vector<complex<double> > dTarget_dF, sizeAdjusted_dTarget_dF;

    //readGradientLargeInp(inputFilePath, dTarget_dF_file_name, test_files);

    readInputFile(inputFilePath, dTarget_dF_FileName, structureFiles, hklFiles, gradientFiles);

    size_t nDataSets = structureFiles.size();


    //############### read dTarget/dF ######################
    string dTarget_dSF_path = makePrefix(customDir, customDirPath, defaultMainDir) + dTarget_dF_FileName;
    test_utilities::read_dTarget_dSF(dTarget_dSF_path, dTarget_dF);
    
    //############### run calculations ######################
    
    r_factors.resize(nDataSets);
    testValidity.resize(nDataSets);

    for( size_t i = 0 ; i<nDataSets ; i++ )
    {
        size_t lastDotPosition = structureFiles[i][0].find_last_of('.');
        string system_name;
        lastDotPosition == string::npos ? system_name = structureFiles[i][0] : 
                                          system_name = structureFiles[i][0].substr(0,lastDotPosition);

        if(mMessageLevel>0)
            cout<< "processing " << system_name << endl;

        r_factors[i].first = system_name; 
        for( size_t j = 0;j< structureFiles[i].size(); j++)
            structureFiles[i][j] = structureFiles[i][j];
        hklFiles[i] = prefixHkl + hklFiles[i];

        gradientFiles[i] = makePrefix(customDir, customDirPath, gradientFilesDir()) + gradientFiles[i];

        calculateR_Factor(customDir, customDirPath, structureFiles[i], hklFiles[i],
                          gradientFiles[i], r_factors[i].second, testValidity[i], dTarget_dF);
    }
    
    //############### check if R-factor is within error threshold

    double threshold = 1e-05;
    
    for (size_t i = 0; i<nDataSets; i++)
    {
        const GradientData &grad_r_factor = r_factors[i].second;
        const GradientData &valid = testValidity[i];
        bool dataSetTestPassed = true;

        if (mMessageLevel>1)
            cout << r_factors[i].first << endl;

        if(grad_r_factor.hasOccupancy)
            checkGradient("occupancy", grad_r_factor.occupancy, valid.occupancy, threshold, dataSetTestPassed, mMessageLevel);

        if(grad_r_factor.hasPosition_cartesian)
            checkGradient("xyz cartesian", grad_r_factor.position_cartesian,testValidity[i].position_cartesian, 3, threshold , dataSetTestPassed, mMessageLevel);

        if (grad_r_factor.hasPosition_fractional)
            checkGradient("xyz fractional", &grad_r_factor.position_fractional[0],valid.position_fractional, 3, threshold, dataSetTestPassed, mMessageLevel);
        
        if (grad_r_factor.has_u_cartesian)
            checkGradient("U_Cartesian", &grad_r_factor.u_cartesian[0], valid.u_cartesian, 6, threshold, dataSetTestPassed, mMessageLevel);
        
        if (grad_r_factor.has_u_cif)
            checkGradient("U_fractional", &grad_r_factor.u_cif[0],valid.u_cif, 6, threshold, dataSetTestPassed, mMessageLevel);
        
        if (grad_r_factor.has_u_star)
            checkGradient("U*", grad_r_factor.u_star,valid.u_star, 6, threshold, dataSetTestPassed, mMessageLevel);
        
        if (grad_r_factor.has_u_iso)
            checkGradient("U_iso", grad_r_factor.u_iso, valid.u_iso, threshold, dataSetTestPassed, mMessageLevel);

        if(dataSetTestPassed)
            setsWhichPassedTest.push_back(r_factors[i].first);
        else
            setsWhichFailedTest.push_back(r_factors[i].first);
    }

    if(mMessageLevel>0)
    {
        cout<< "data sets which passed the test:" << endl;
        for(size_t i = 0;i<setsWhichPassedTest.size();i++)
            cout<< "   " << setsWhichPassedTest[i] << endl;

        cout << "data sets which failed the test:" << endl;
        for (size_t i = 0; i<setsWhichFailedTest.size(); i++)
            cout<< "   " << setsWhichFailedTest[i] << endl;
    }

    return setsWhichFailedTest.empty();
}

void TestGradientLarge::checkGradient(
    const string &parameterType, 
    const double * const gradient,
    const double * const validity,
    size_t nComponents,
    double threshold, 
    bool &testState, 
    int messageLevel)
{
    cout << setprecision(8);
    size_t i;
    for(i=0;i<nComponents;i++)
        testState = testState && (gradient[i]<threshold);
    
    if(messageLevel>1)
    {
        cout<< "    " << parameterType << "  ";
        for(i=0;i<nComponents;i++)
            cout<< " " << gradient[i];
        cout<< endl;
        cout<< "        valid test " ;
        for (i = 0; i<nComponents; i++)
            cout<< boolalpha << (validity[i]!=0.0) << "  ";
        cout<< endl;
    }
}

void TestGradientLarge::checkGradient(
    const string &parameterType, 
    double gradient, 
    double validity,
    double threshold, 
    bool &testState,
    int messageLevel)
{
    cout<< setprecision(8);
    if (messageLevel>1)
    {
        cout << "    " << parameterType << "  " << gradient << endl;
        cout << "        valid test " << boolalpha << (validity != 0.0) << endl;
    }
    testState = testState && (gradient < threshold);
}

bool TestGradientLarge::hasNonzeroElement(
    const std::vector<double> &v)
{
    //return !( *min_element(v.begin(),v.end())==0.0 && *max_element(v.begin(),v.end())==0.0);
    double threshold = 1e-8;
    double minValue = *min_element(v.begin(), v.end());
    double maxValue = *max_element(v.begin(), v.end());
    return ( fabs(minValue)>threshold || fabs(maxValue)>threshold );
}

void TestGradientLarge::calculateR_Factor(
    std::vector<double> &data, 
    std::vector<double> &referenceData,
    double &rFactor, 
    double &validData,
    bool &hasData)
{
    if ( (!data.empty()) && (!referenceData.empty()) )
    {
        hasData = true;

        if (!hasNonzeroElement(referenceData))
        {
            rFactor = 0;
            validData = 0;
        }
        else
        {
            rFactor = test_utilities::calculateR_Factor(data, referenceData, false);
            validData = 1;
        }

        data.clear();
        referenceData.clear();
    }

}

void TestGradientLarge::calculateR_Factor(
    bool customDir,
    const std::string &customDirPath,
    const std::vector<std::string> &structureFiles,
    const std::string &hkl_file,
    const std::string &gradients_file,
    GradientData &gradientR_Factor,
    GradientData &validData,
    const vector<complex<double> > &dTarget_dF)
{
    vector<Vector3i> hkl;
    vector<complex<double> > ref_f, f;
    //vector<complex<double> > f;
    vector<TargetFunctionAtomicParamDerivatives> dTarg_dParam;
    Crystal crystal;

    hkl_io::readCompactHklFile(hkl_file,hkl);

    vector<GradientData> calcGradientData, referenceGradientData;
    size_t atomIdx,nAtoms,i;
    

    readGradientFile(gradients_file, referenceGradientData);
    
    vector<complex<double> > sizeAjusted_dTarget_dF;
    sizeAjusted_dTarget_dF = dTarget_dF;
    sizeAjusted_dTarget_dF.resize(hkl.size());

    readStructureAndCalculateDerivatives(customDir, customDirPath, structureFiles, hkl, sizeAjusted_dTarget_dF, crystal, dTarg_dParam);

    
    //########### Assign calculated gradient data #####################

    nAtoms = dTarg_dParam.size();
    calcGradientData.resize(nAtoms);
    for(atomIdx=0 ; atomIdx<nAtoms ; atomIdx++)
    {
        calcGradientData[atomIdx].occupancy = dTarg_dParam[atomIdx].occupancy_derivatives;
        calcGradientData[atomIdx].hasOccupancy = true;
        
        size_t nAdpComponents = dTarg_dParam[atomIdx].adp_derivatives.size();

        if(nAdpComponents == 1)
        {
            calcGradientData[atomIdx].u_iso = dTarg_dParam[atomIdx].adp_derivatives[0];
            calcGradientData[atomIdx].has_u_iso = true;
        }
        
        if(nAdpComponents == 6)
        {
            for(i=0;i<6;i++)
                calcGradientData[atomIdx].u_cartesian[i] = dTarg_dParam[atomIdx].adp_derivatives[i];
            calcGradientData[atomIdx].has_u_cartesian = true;
        }

        calcGradientData[atomIdx].hasPosition_cartesian = true;
        for(i=0;i<3;i++)
            calcGradientData[atomIdx].position_cartesian[i] =  dTarg_dParam[atomIdx].atomic_position_derivatives[i];

    }
    //######### calculate transformed gradients

    calculateTransformedGradients(crystal.unitCell, calcGradientData);

    //######### calculate R-factors for gradients ##############

    gradientR_Factor.clear();
    vector<vector<double> > calculatedData(6), referenceData(6);
    // occupancy

    for(atomIdx=0; atomIdx<nAtoms;atomIdx++)
        if(calcGradientData[atomIdx].hasOccupancy && referenceGradientData[atomIdx].hasOccupancy)
        {
            calculatedData[0].push_back(calcGradientData[atomIdx].occupancy);
            referenceData[0].push_back(referenceGradientData[atomIdx].occupancy);
        }


     calculateR_Factor(calculatedData[0], referenceData[0], gradientR_Factor.occupancy, 
                      validData.occupancy, gradientR_Factor.hasOccupancy);

    // position cartesian

    for (atomIdx = 0; atomIdx<nAtoms; atomIdx++)
        if (calcGradientData[atomIdx].hasPosition_cartesian && referenceGradientData[atomIdx].hasPosition_cartesian)
            for(i=0;i<3;i++)
            {
                calculatedData[i].push_back(calcGradientData[atomIdx].position_cartesian[i]);
                referenceData[i].push_back(referenceGradientData[atomIdx].position_cartesian[i]);
            }

    for (i = 0; i<3; i++)
        calculateR_Factor(calculatedData[i], referenceData[i], gradientR_Factor.position_cartesian[i],
                          validData.position_cartesian[i], gradientR_Factor.hasPosition_cartesian);


    // position fractional

    for (atomIdx = 0; atomIdx<nAtoms; atomIdx++)
        if (calcGradientData[atomIdx].hasPosition_fractional && referenceGradientData[atomIdx].hasPosition_fractional)
            for (i = 0; i<3; i++)
            {
                calculatedData[i].push_back(calcGradientData[atomIdx].position_fractional[i]);
                referenceData[i].push_back(referenceGradientData[atomIdx].position_fractional[i]);
            }

    for (i = 0; i<3; i++)
        calculateR_Factor(calculatedData[i], referenceData[i], gradientR_Factor.position_fractional[i],
                          validData.position_cartesian[i], gradientR_Factor.hasPosition_fractional);


    // u_iso

    for (atomIdx = 0; atomIdx<nAtoms; atomIdx++)
        if (calcGradientData[atomIdx].has_u_iso && referenceGradientData[atomIdx].has_u_iso)
        {
            calculatedData[0].push_back(calcGradientData[atomIdx].u_iso);
            referenceData[0].push_back(referenceGradientData[atomIdx].u_iso);
        }

    calculateR_Factor(calculatedData[0], referenceData[0], gradientR_Factor.u_iso,
                      validData.u_iso, gradientR_Factor.has_u_iso);


    // u_cartesian

    for (atomIdx = 0; atomIdx<nAtoms; atomIdx++)
        if (calcGradientData[atomIdx].has_u_cartesian && referenceGradientData[atomIdx].has_u_cartesian)
            for (i = 0; i<6; i++)
            {
                calculatedData[i].push_back(calcGradientData[atomIdx].u_cartesian[i]);
                referenceData[i].push_back(referenceGradientData[atomIdx].u_cartesian[i]);
            }

    for (i = 0; i<6; i++)
        calculateR_Factor(calculatedData[i], referenceData[i], gradientR_Factor.u_cartesian[i],
                          validData.u_cartesian[i], gradientR_Factor.has_u_cartesian);

    // u star

    for (atomIdx = 0; atomIdx<nAtoms; atomIdx++)
        if (calcGradientData[atomIdx].has_u_star && referenceGradientData[atomIdx].has_u_star)
            for (i = 0; i<6; i++)
            {
                calculatedData[i].push_back(calcGradientData[atomIdx].u_star[i]);
                referenceData[i].push_back(referenceGradientData[atomIdx].u_star[i]);
            }

    for (i = 0; i<6; i++)
        calculateR_Factor(calculatedData[i], referenceData[i], gradientR_Factor.u_star[i],
                          validData.u_star[i], gradientR_Factor.has_u_star);


}

void TestGradientLarge::calculateTransformedGradients(
    const UnitCell &unitCell, 
    std::vector<GradientData> &gradientData)
{
    StructuralParametersConverter derivativesConverter;
    size_t i, n = gradientData.size();
    Vector3<complex<double> > v3cIn,v3cOut;
    vector<complex<double> > v6cIn(6),v6cOut(6);

    derivativesConverter.set(unitCell);

    for (i = 0; i<n; i++)
    {
        if (gradientData[i].hasPosition_cartesian)
        {
            v3cIn = gradientData[i].position_cartesian;
            derivativesConverter.derivativesXyzCartesianToFractional(v3cIn,v3cOut);
            gradientData[i].hasPosition_fractional = true;
            for (int j = 0; j<3; j++)
                gradientData[i].position_fractional[j] = v3cOut[j].real();
        }

        if (gradientData[i].has_u_cartesian)
        {
            for(int j=0;j<6;j++)
                v6cIn[j] = gradientData[i].u_cartesian[j];
            derivativesConverter.derivativesU_CartesianToU_Star(v6cIn, v6cOut);
            for (int j = 0; j<6; j++)
                gradientData[i].u_star[j] = v6cOut[j].real();
            gradientData[i].has_u_star = true;
        }

        if (gradientData[i].has_u_star)
        {
            for (int j = 0; j<6; j++)
                v6cIn[j] = gradientData[i].u_star[j];
            derivativesConverter.derivativesU_StarToU_Cif(v6cIn, v6cOut);
            for (int j = 0; j<6; j++)
                gradientData[i].u_cif[j] = v6cOut[j].real();
            gradientData[i].has_u_cif = true;
        }
    }

}

void TestGradientLarge::readGradientFile(
    const std::string &fileName, 
    std::vector<GradientData> &gradientData)
{
    vector<vector<string> > splittedLines;
    ifstream in(fileName.c_str());
    string line;
    vector<string> splittedLine;
    

    // read input data as splitted lines

    if(!in.good())
        on_error::throwException(string("cannot read file '") + fileName+ string("with gradient data"),
                                 __FILE__, __LINE__);
    while(in.good())
    {
        getline(in,line);
        string_utilities::split(line, splittedLine);
        if(!splittedLine.empty())
            splittedLines.push_back(splittedLine);
    }
    in.close();

    // process read input data
    GradientData data;
    for(size_t i=0,n=splittedLines.size();i<n;i++)
    {
        if(splittedLines[i].size()==1 && i>1)
        {
            gradientData.push_back(data);
            data.clear();
        }

        if (splittedLines[i][0] == string("occ"))
        {
            data.hasOccupancy = true;
            data.occupancy = atof(splittedLines[i][1].c_str());
        }
        if ( splittedLines[i][0] == string("xyz_Cart") )
        {
            data.hasPosition_cartesian = true;
            for(int j=0;j<3;j++)
                data.position_cartesian[j] = atof(splittedLines[i][j+1].c_str());
        }
        if ( splittedLines[i][0] == string("xyz_fract") )
        {
            data.hasPosition_fractional = true;
            for (int j = 0; j<3; j++)
                data.position_fractional[j] = atof(splittedLines[i][j + 1].c_str());
        }
        if ( splittedLines[i][0] == string("U_cart") )
        {
            data.has_u_cartesian = true;
            for (int j = 0; j<6; j++)
                data.u_cartesian[j] = atof(splittedLines[i][j + 1].c_str());
        }
        if ( splittedLines[i][0] == string("U_cif") )
        {
            data.has_u_cif = true;
            for (int j = 0; j<6; j++)
                data.u_cif[j] = atof(splittedLines[i][j + 1].c_str());
        }
        if (splittedLines[i][0] == string("U_star"))
        {
            data.has_u_star = true;
            for (int j = 0; j<6; j++)
                data.u_star[j] = atof(splittedLines[i][j + 1].c_str());
        }
        if (splittedLines[i][0] == string("U_iso"))
        {
            data.has_u_iso = true;
            data.u_iso = atof(splittedLines[i][1].c_str());
        }

    }
    gradientData.push_back(data);
}

void TestGradientLarge::readGradientLargeInp(
    const std::string & inputFile, 
    std::string & dTargetdParameterFile, 
    std::vector<std::vector<std::string> >& dataFiles)
{
    ifstream in(inputFile.c_str());
    string s, line;
    vector<string> words;
    if (!in.good())
        on_error::throwException(string("can not open input file: '") + inputFile +
                                 string("' when testing gradients"), __FILE__, __LINE__);

    in >> s >> s >> dTargetdParameterFile;
    in >> s >> s >> s;
    size_t nDataSets = atoi(s.c_str());

    getline(in, line);

    for (size_t i = 0; i<nDataSets; i++)
    {
        if (!in.good())
            on_error::throwException(string("unexpected end of input file: '") + inputFile +
                                     string("' when testing gradients"), __FILE__, __LINE__);
        getline(in, line);
        string_utilities::split(line, words);
        if (words.size() != 3)
            on_error::throwException(string("invalid format of input file: '") + inputFile +
                                     string("' when testing gradients"), __FILE__, __LINE__);
        dataFiles.push_back(words);
    }

    in.close();
}

void TestGradientLarge::writeResults(
    const std::string &xdMasFileName,
    const std::vector<AtomInCrystal> &atoms,
    const std::vector<Vector3i> &hkl,
    const std::vector<std::complex<double> > &f,
    std::vector<TargetFunctionAtomicParamDerivatives> &dTarg_dParam)
{
    // filename
    string fileNameCore, fileName = xdMasFileName;

    if (fileName.find_last_of('\\') != string::npos)
        fileName = fileName.substr(fileName.find_last_of('\\') + 1);
    if (fileName.find_last_of('/') != string::npos)
        fileName = fileName.substr(fileName.find_last_of('/') + 1);

    if (fileName.size()>7)
    {
        if (fileName.substr(fileName.size() - 7) == string("_xd.mas"))
            fileNameCore = fileName.substr(0, fileName.size() - 7);
        else
            fileNameCore = fileName.substr(0, fileName.size() - 4);
    }
    else
        fileNameCore = fileName.substr(0, fileName.size() - 4);

    //fileNameCore = fileNameCore.substr()
    // write structure factors
    ofstream outSF((fileNameCore + string(".sf")).c_str());
    size_t hklIndex, nHkl = hkl.size();

    if (!outSF.good())
        on_error::throwException("can not write structure factor file for saving MultipolarGradientLarge results", __FILE__, __LINE__);

    outSF << "number of reflections " << nHkl << endl;

    outSF << setprecision(12);

    for (hklIndex = 0; hklIndex<nHkl; hklIndex++)
        outSF << hkl[hklIndex].x << " " << hkl[hklIndex].y << " " << hkl[hklIndex].z
        << " " << f[hklIndex].real() << " " << f[hklIndex].real() << endl;

    outSF.close();

    // write gradient data

    ofstream out((fileNameCore + string(".grd")).c_str());
    size_t atomIdx, nAtoms = atoms.size();
    out << setprecision(12);

    for (atomIdx = 0; atomIdx < nAtoms; atomIdx++)
    {
        out << atoms[atomIdx].label << endl;
        out << "   occ " << dTarg_dParam[atomIdx].occupancy_derivatives << endl;
        out << "   xyz_Cart ";
        for (int i = 0; i<3; i++)
            out << dTarg_dParam[atomIdx].atomic_position_derivatives[i] << " ";
        out << endl;

        if (dTarg_dParam[atomIdx].adp_derivatives.size() == 6)
        {
            out << "   U_cart";
            for (int i = 0; i<6; i++)
                out << " " << dTarg_dParam[atomIdx].adp_derivatives[i];
            out << endl;
        }
        if (dTarg_dParam[atomIdx].adp_derivatives.size() == 1)
            out << "   U_iso " << dTarg_dParam[atomIdx].adp_derivatives[0] << endl;
    }
    out.close();
    /*
    C5
    occ -67.4669057019
    xyz_Cart 40.4718571689 16.2773418083 -73.0288589069
    xyz_fract 152.170135769 149.959866749 -497.921767298
    U_cart 1131.70945227 978.643707882 608.826246635 163.357076786 409.532898863 -26.4967138597
    H5
    occ -2.49240562667
    xyz_Cart 9.43867663863 7.36064219201 0.781785558247
    xyz_fract 35.4884802936 55.2565133458 10.7021380331
    U_iso 10.4642879987
    */

}
