#include "Test_GradientSmall.h"
#include "test_utilities.h"
#include "data_directory.h"

#include "discamb/MathUtilities/Vector3.h"

#include "discamb/BasicUtilities/OnError.h"
#include "discamb/BasicUtilities/StringUtilities.h"
#include "discamb/CrystalStructure/ReciprocalLatticeUnitCell.h"

#include "discamb/CrystalStructure/StructuralParametersConverter.h"

#include "discamb/IO/NativeIAM_Reader.h"


#include <fstream>

using namespace std;
using namespace discamb;



TestGradientSmall::TestGradientSmall()
{
}

TestGradientSmall::~TestGradientSmall()
{
}

void TestGradientSmall::readGradientSmallInp(
    const std::string &inputFile,
    std::vector<std::vector<std::string> > &structureFiles,
    std::vector<std::string> &gradientFiles)
{
    structureFiles.clear();
    gradientFiles.clear();

    ifstream in(inputFile.c_str());
    string s, line;
    vector<string> words;
    if (!in.good())
        on_error::throwException(string("can not open input file: '") + inputFile +
                                 string("' when testing gradients"), __FILE__, __LINE__);

    in >> s >> s >> s;
    size_t nDataSets = atoi(s.c_str());

    getline(in, line);

    for (size_t i = 0; i<nDataSets; i++)
    {
        if (!in.good())
            on_error::throwException(string("unexpected end of input file: '") + inputFile +
                                     string("' when testing gradients"), __FILE__, __LINE__);
        getline(in, line);
        string_utilities::split(line, words);
        if (words.size() < 2)
            on_error::throwException(string("invalid format of input file: '") + inputFile +
                                     string("' when testing gradients"), __FILE__, __LINE__);
        structureFiles.push_back(vector<string>(words.begin(),words.end()-1));
        gradientFiles.push_back(words.back());
    }

    in.close();

}


bool TestGradientSmall::run_test(
    std::string &message)
{
    message.clear();


    string mainDir = correcness_test_1_data_directory;

    string inputFilePath = mainDir + string("/") + inputFileName();//string("test_iam_gradient_small.inp");
    
    string s,dTarget_dF_file_name, line;
    vector<GradientRFactorSM> rFactorData;
    vector<string> setsWhichPassedTest,setsWhichFailedTest;

    /* testValidity[i].parameter_value is set to 1 
       if the dataset for the parameter contains numbers effectively different than 0
       (i.e. fabs(data[i]) > threshold, threshold defined in the hasNonzeroElement method)
       otherwise it is set to 0 (and valid R-factor test cannot be performed */


    vector<string> words,gradientFiles;
    vector<vector<string> > structureFiles;
    
    readGradientSmallInp(inputFilePath, structureFiles, gradientFiles);
                        

    size_t nDataSets = structureFiles.size();


    
    //############### run calculations ######################
    rFactorData.resize(nDataSets);
    vector<double> sfR_Factors(nDataSets);

    for( size_t i = 0 ; i<nDataSets ; i++ )
    {
        string systemName = test_utilities::structureNameFromFileName(structureFiles[i][0]);
        
        if(mMessageLevel>0)
            cout<< "processing " << systemName << endl;
    
        for(size_t j=0;j<structureFiles[i].size();j++)
            structureFiles[i][j] = structureFiles[i][j];
        gradientFiles[i] = gradientFiles[i];

        sfR_Factors[i] = calculateR_Factor(structureFiles[i], gradientFiles[i], rFactorData[i]);
        rFactorData[i].systemLabel = systemName;
    }
    
    //############### check if R-factor is within error threshold

    double threshold = 1e-6;

    for (size_t i = 0; i<nDataSets; i++)
    {
        if( rFactorTestPassed(rFactorData[i], mMessageLevel) && sfR_Factors[i]<threshold)
            setsWhichPassedTest.push_back(rFactorData[i].systemLabel);
        else
            setsWhichFailedTest.push_back(rFactorData[i].systemLabel);
        if(mMessageLevel>1)
            cout<< "r factor " << sfR_Factors[i] << endl;
    }


    if(mMessageLevel>0)
    {
        cout<< "data sets which passed the test:" << endl;
        for(size_t i = 0;i<setsWhichPassedTest.size();i++)
            cout<< "   " << setsWhichPassedTest[i] << endl;

        cout << "data sets which failed the test:" << endl;
        for (size_t i = 0; i<setsWhichFailedTest.size(); i++)
            cout<< "   " << setsWhichFailedTest[i] << endl;
    }

    return setsWhichFailedTest.empty();
}


void TestGradientSmall::checkGradient(
    const string &parameterType, 
    const double * const gradient,
    const double * const validity,
    size_t nComponents,
    double threshold, 
    bool &testState, 
    int messageLevel)
{
    size_t i;
    for(i=0;i<nComponents;i++)
        testState = testState && (gradient[i]<threshold);
    
    if(messageLevel>1)
    {
        cout<< "    " << parameterType << "  ";
        for(i=0;i<nComponents;i++)
            cout<< " " << gradient[i];
        cout<< endl;
        cout<< "        valid test " ;
        for (i = 0; i<nComponents; i++)
            cout<< boolalpha << (validity[i]!=0.0) << "  ";
        cout<< endl;
    }
}

void TestGradientSmall::checkGradient(
    const string &parameterType, 
    double gradient, 
    double validity,
    double threshold, 
    bool &testState,
    int messageLevel)
{
    if (messageLevel>1)
    {
        cout << "    " << parameterType << "  " << gradient << endl;
        cout << "        valid test " << boolalpha << (validity != 0.0) << endl;
    }
    testState = testState && (gradient < threshold);
}

bool TestGradientSmall::hasNonzeroElement(
    const std::vector<double> &v)
{
    
    double threshold = 1e-8;
    double minValue = *min_element(v.begin(), v.end());
    double maxValue = *max_element(v.begin(), v.end());
    return ( fabs(minValue)>threshold || fabs(maxValue)>threshold );
}



/*
void TestGradientSmall::calculateTransformedGradients(
    const UnitCell &unitCell, 
    GradientDataSM &gradientData)
{
    StructuralParametersConverter derivativesConverter;
    derivativesConverter.set(unitCell);
    size_t atomIdx, nAtoms, nHkl, i; 
    Vector3<complex<double> > vec3cIn,vec3cOut;
    vector<complex<double> > vec6cIn(6),vec6cOut(6);
    

    nAtoms = gradientData.atomLabel.size();
    
    if(nAtoms==0)
        return;

    nHkl = gradientData.occupancy[0].size();

    for(atomIdx = 0 ; atomIdx<nAtoms; atomIdx++)
    {
        if( !gradientData.hasXyzCart && gradientData.hasXyzFract)
        {
            for(size_t hklIdx=0;hklIdx<nHkl;hklIdx++)
            {
                for(i=0;i<3;i++)
                    vec3cIn[i] = gradientData.position_fractional[atomIdx][i][hklIdx]; 
                derivativesConverter.derivativesXyzFractionalToCartesian(vec3cIn, vec3cOut);
            }
        }
        if (gradientData.hasXyzCart && !gradientData.hasXyzFract)
        {
            for (size_t hklIdx = 0; hklIdx<nHkl; hklIdx++)
            {
                for (i = 0; i<3; i++)
                    vec3cIn[i] = gradientData.position_cartesian[atomIdx][i][hklIdx];
                derivativesConverter.derivativesXyzCartesianToFractional(vec3cIn, vec3cOut);
                
                for (i = 0; i<3; i++)
                    gradientData.position_fractional[atomIdx][i][hklIdx] = vec3cOut[i];
            }
        }
        
        if( gradientData.hasUcart && gradientData.u_cartesian[atomIdx].size() == 6 )
        {
            
            if(!gradientData.hasUstar)
                convertAdps( gradientData.u_cartesian[atomIdx], gradientData.u_star[atomIdx], 
                             &StructuralParametersConverter::derivativesU_CartesianToU_Star, derivativesConverter);

            if (!gradientData.hasUcif)
                convertAdps(gradientData.u_cartesian[atomIdx], gradientData.u_cif[atomIdx],
                            &StructuralParametersConverter::derivativesU_CartesianToU_Cif, derivativesConverter);
        }

        if (gradientData.hasUstar && gradientData.u_star[atomIdx].size() == 6)
        {

            if (!gradientData.hasUcart)
                convertAdps(gradientData.u_star[atomIdx] , gradientData.u_cartesian[atomIdx],
                            &StructuralParametersConverter::derivativesU_StarToU_Cartesian, derivativesConverter);

            if (!gradientData.hasUcif)
                convertAdps(gradientData.u_star[atomIdx], gradientData.u_cif[atomIdx],
                            &StructuralParametersConverter::derivativesU_StarToU_Cif, derivativesConverter);
        }

        if (gradientData.hasUcif && gradientData.u_cif[atomIdx].size() == 6)
        {

            if (!gradientData.hasUstar)
                convertAdps(gradientData.u_cif[atomIdx], gradientData.u_star[atomIdx],
                            &StructuralParametersConverter::derivativesU_CifToU_Star, derivativesConverter);

            if (!gradientData.hasUcart)
                convertAdps(gradientData.u_cif[atomIdx], gradientData.u_cartesian[atomIdx],
                            &StructuralParametersConverter::derivativesU_CifToU_Cartesian, derivativesConverter);
        }

    
    }


}*/

void TestGradientSmall::readGradientAndHkl(
    const std::string &fileName, 
    std::vector<std::complex<double> > &sf,
    GradientDataSM_AllConventions &gradientData,
    std::vector<Vector3i> &hkl)
{
    // clear data
    gradientData.clear();
    hkl.clear();
    sf.clear();

    ifstream in(fileName.c_str());

    if(!in.good())
        on_error::throwException(string("can not read gradient and hkl data file: '")
                                 + fileName + string("' when testing gradient calculation, version for small molecules"),
                                 __FILE__,__LINE__);

    vector<string> words;
    string line;
    Vector3i hklIndex;
    int atomIndex, i;
    complex<double> value;

    atomIndex=-1;//to silent compiler warning

    while(in.good())
    {
        getline(in,line);
        string_utilities::split(line,words);
        double real,imaginary;
        if(words.empty())
            continue;
        if(words[0] == string("hkl"))
        {
            atomIndex = -1;
            hklIndex.x = atoi(words[2].substr(1, words[2].size() - 2).c_str() );
            hklIndex.y = atoi(words[3].substr(0, words[3].size() - 1).c_str() );
            hklIndex.z = atoi(words[4].substr(0, words[4].size() - 1).c_str() );
            hkl.push_back(hklIndex);
            sf.push_back(complex<REAL>(atof(words[6].c_str()), atof(words[7].c_str())));
        }
        else if(words.size()==1)
        {
            atomIndex++;

            if(hkl.size()==1)
            {
                int nAtoms = atomIndex + 1;
                gradientData.atomLabel.push_back(words[0]);
                gradientData.occupancy.resize(nAtoms);
                gradientData.position_cartesian.resize(nAtoms,vector<vector<complex<double> > >(3));
                gradientData.position_fractional.resize(nAtoms,vector<vector<complex<double> > >(3));
                gradientData.u_cartesian.resize(nAtoms, vector<vector<complex<double> > >(6));
                gradientData.u_iso.resize(nAtoms);
                gradientData.u_star.resize(nAtoms, vector<vector<complex<double> > >(6));
            }
        }
        else if(words[0] == string("occ"))
        {
            real = atof(words[1].c_str());
            imaginary = atof(words[2].c_str());
            value = complex<double>(real,imaginary);
            gradientData.occupancy[atomIndex].push_back(value);
        }
        else if(words[0] == string("xyz_Cart"))
        {
            for(i=0;i<3;i++)
                gradientData.position_cartesian[atomIndex][i].push_back(atof(words[i + 1].c_str()));
        }
        else if (words[0] == string("xyz_fract"))
        {
            for (i = 0; i<3; i++)
            {
                real = atof(words[2*i+1].c_str());
                imaginary = atof(words[2*i+2].c_str());
                value = complex<double>(real, imaginary);
                gradientData.position_fractional[atomIndex][i].push_back(value);
            }
        }
        else if (words[0] == string("U_cart"))
        {
            for (i = 0; i<6; i++)
                gradientData.u_cartesian[atomIndex][i].push_back( atof(words[i + 1].c_str()) );
        }
        else if (words[0] == string("U_star"))
        {
            for (i = 0; i<6; i++)
            {
                real = atof(words[2 * i + 1].c_str());
                imaginary = atof(words[2 * i + 2].c_str());
                value = complex<double>(real, imaginary);
                gradientData.u_star[atomIndex][i].push_back(value);
            }
        }
        else if (words[0] == string("U_iso"))
        {
            real = atof(words[1].c_str());
            imaginary = atof(words[2].c_str());
            value = complex<double>(real, imaginary);
            gradientData.u_iso[atomIndex].push_back(value);
        }
        else
            on_error::throwException(string("error when processing gradient and hkl data file: '")
                                     + fileName + string("' when testing gradient calculation, version for small molecules"),
                                     __FILE__, __LINE__);

    }

    in.close();

}

/*
void GradientDataSM::allocate(
    size_t nHkl, 
    const std::vector<size_t> &nAdpsPerAtom)
{
    size_t atomIdx,nAtoms = nAdpsPerAtom.size();

    clear();
    atomLabel.resize(nAtoms);
    occupancy.resize(nAtoms,vector<complex<double> >(nHkl));
    position_fractional.resize(nAtoms, vector<vector<complex<double> > >(3,vector<complex<double> >(nHkl)));
    position_cartesian.resize(nAtoms, vector<vector<complex<double> > >(3, vector<complex<double> >(nHkl)));

    u_iso.resize(nAtoms);
    u_cartesian.resize(nAtoms);
    u_star.resize(nAtoms);
    u_cif.resize(nAtoms);
    
    for(atomIdx=0;atomIdx<nAtoms;atomIdx++)
        if ( nAdpsPerAtom[atomIdx] == 1 )
        {
            u_iso[atomIdx].resize(nHkl);
        }
        else if (nAdpsPerAtom[atomIdx] == 6)
        {
            u_cartesian[atomIdx].resize(6,vector<complex<double> >(nHkl));
            u_star[atomIdx].resize(6, vector<complex<double> >(nHkl));
            u_cif[atomIdx].resize(6, vector<complex<double> >(nHkl));
        }
}

void GradientDataSM::clear()
{
    hasXyzCart = hasXyzFract = hasUcart = hasUstar = hasUcif = false;
    atomLabel.clear();
    occupancy.clear();
    u_iso.clear();
    position_fractional.clear();
    position_cartesian.clear();
    u_cartesian.clear();
    u_star.clear();
}
*/
/*
void TestGradientSmall::convertGradientToGenerateMissingData(
    GradientDataSM &gradientData,
    const UnitCell &unitCell)
{
    size_t nAtoms, atomIdx, nHkl, hklIdx;
    StructuralParametersConverter derivativesConverter;
    Vector3<complex<double> > derivativePositionCart, derivativePositionFract;
    vector<complex<double> > uIn(6),uOut(6);

    nAtoms = gradientData.occupancy.size();

    if (nAtoms == 0)
        return;

    derivativesConverter.set(unitCell);
    nHkl = gradientData.occupancy[0].size();
    if (!gradientData.hasXyzCart && gradientData.hasXyzFract)
        for (atomIdx = 0; atomIdx < nAtoms; atomIdx++)
        {
            gradientData.position_cartesian[atomIdx].assign(3, vector<complex<double> >(nHkl, 0.0));
            for (hklIdx = 0; hklIdx < nHkl; hklIdx++)
            {
                for(int i=0;i<3;i++)
                    derivativePositionFract[i] = gradientData.position_fractional[atomIdx][i][hklIdx];
                derivativesConverter.derivativesXyzFractionalToCartesian(derivativePositionFract, derivativePositionCart);
                for (int i = 0; i<3; i++)
                    gradientData.position_cartesian[atomIdx][i][hklIdx] = derivativePositionCart[i] ;
            }
        }

    if (gradientData.hasXyzCart && !gradientData.hasXyzFract)
        for (atomIdx = 0; atomIdx < nAtoms; atomIdx++)
        {
            gradientData.position_fractional[atomIdx].assign(3, vector<complex<double> >(nHkl, 0.0));
            for (hklIdx = 0; hklIdx < nHkl; hklIdx++)
            {
                for (int i = 0; i<3; i++)
                    derivativePositionCart[i] = gradientData.position_cartesian[atomIdx][i][hklIdx];
                derivativesConverter.derivativesXyzCartesianToFractional(derivativePositionCart, derivativePositionFract );
                for (int i = 0; i<3; i++)
                    gradientData.position_fractional[atomIdx][i][hklIdx] = derivativePositionFract[i];

            }
        }

    

    if(gradientData.hasUcart)
    {
        if(!gradientData.hasUcif)
        {
            gradientData.u_cif.clear();
            gradientData.u_cif.resize(nAtoms);

            for (atomIdx = 0; atomIdx < nAtoms; atomIdx++)
                if(!gradientData.u_cartesian[atomIdx].empty())
                {
                    gradientData.u_cif[atomIdx].resize(6,vector<complex<double> >(nHkl));

                    for (hklIdx = 0; hklIdx < nHkl; hklIdx++)
                    {
                        for (int i = 0; i<6; i++)
                            uIn[i] = gradientData.u_cartesian[atomIdx][i][hklIdx];
                        derivativesConverter.derivativesU_CartesianToU_Cif(uIn,uOut);
                        for (int i = 0; i<6; i++)
                            gradientData.u_cif[atomIdx][i][hklIdx] = uOut[i];
                    }
                    gradientData.hasUcif = true;
                }
            
        }


        if(!gradientData.hasUstar)
        {
            gradientData.u_star.clear();
            gradientData.u_star.resize(nAtoms);

            for (atomIdx = 0; atomIdx < nAtoms; atomIdx++)
                if (!gradientData.u_cartesian[atomIdx].empty())
                {
                    gradientData.u_star[atomIdx].resize(6, vector<complex<double> >(nHkl));

                    for (hklIdx = 0; hklIdx < nHkl; hklIdx++)
                    {
                        for (int i = 0; i<6; i++)
                            uIn[i] = gradientData.u_cartesian[atomIdx][i][hklIdx];
                        derivativesConverter.derivativesU_CartesianToU_Star(uIn, uOut);
                        for (int i = 0; i<6; i++)
                            gradientData.u_star[atomIdx][i][hklIdx] = uOut[i];
                    }
                    gradientData.hasUstar = true;
                }
        }
    } // gradientData.hasUcart

}
*/

double TestGradientSmall::calculateR_Factor(
    const std::vector<std::string> &structureFiles,
    const std::string &gradientsFile,
    GradientRFactorSM &gradientR_Factor)
{
    GradientDataSM_AllConventions refGradientData, calcGradientData;
    //GradientDataSM calcGradientData;
    vector<complex<REAL> > f_ref, f_calc;
    vector<Vector3i> hkl;
    string mainDir = correcness_test_1_data_directory;
    readGradientAndHkl(mainDir + string("/") + hklAndGradientDirName() + string("/") + gradientsFile, f_ref, refGradientData, hkl);
    calculateGradientData(structureFiles, hkl, f_calc, calcGradientData);

    size_t nAtoms, atomIdx, nHkl, hklIndex, i;
    nHkl = hkl.size();
    
    nAtoms = refGradientData.occupancy.size();
    
    
    gradientR_Factor.occupancy.resize(nAtoms);
    gradientR_Factor.position_cartesian.resize(nAtoms, vector<double>(3));
    gradientR_Factor.position_fractional.resize(nAtoms, vector<double>(3));
    gradientR_Factor.u_cartesian.resize(nAtoms);
    gradientR_Factor.u_iso.resize(nAtoms);
    gradientR_Factor.u_star.resize(nAtoms);

    gradientR_Factor.atomLabel = refGradientData.atomLabel;

    for( atomIdx=0; atomIdx<nAtoms; atomIdx++)
    {
        
        // occupancy
        gradientR_Factor.occupancy[atomIdx] = test_utilities::calculateR_Factor(calcGradientData.occupancy[atomIdx],
                                                                                refGradientData.occupancy[atomIdx], 
                                                                                false);

        // position cartesian

        for( i=0 ; i<3 ; i++ )
        {
            for(hklIndex=0;hklIndex<nHkl;hklIndex++)
                calcGradientData.position_cartesian[atomIdx][i][hklIndex] = 
                    calcGradientData.position_cartesian[atomIdx][i][hklIndex].real();
            gradientR_Factor.position_cartesian[atomIdx][i] = 
                    test_utilities::calculateR_Factor(calcGradientData.position_cartesian[atomIdx][i],
                                                      refGradientData.position_cartesian[atomIdx][i],
                                                      false);
        }

        // position fractional
        for (i = 0; i<3; i++)
            gradientR_Factor.position_fractional[atomIdx][i] =
                test_utilities::calculateR_Factor(calcGradientData.position_fractional[atomIdx][i],
                                                  refGradientData.position_fractional[atomIdx][i],
                                                  false);
        // U_cartesian
        if( ! calcGradientData.u_cartesian[atomIdx].empty() && 
            ! refGradientData.u_cartesian[atomIdx].empty() )
        {
            gradientR_Factor.u_cartesian[atomIdx].resize(6);
            for (i = 0; i<6; i++)
            {
                for (hklIndex = 0; hklIndex<nHkl; hklIndex++)
                    calcGradientData.u_cartesian[atomIdx][i][hklIndex] =
                    calcGradientData.u_cartesian[atomIdx][i][hklIndex].real();


                gradientR_Factor.u_cartesian[atomIdx][i] = 
                    test_utilities::calculateR_Factor(calcGradientData.u_cartesian[atomIdx][i],
                                                      refGradientData.u_cartesian[atomIdx][i],
                                                      false);
            }
        }
        // U*
        if( !calcGradientData.u_star[atomIdx].empty() && 
            !refGradientData.u_star[atomIdx].empty() )
        {
            gradientR_Factor.u_star[atomIdx].resize(6);
            for (i = 0; i<6; i++)
                gradientR_Factor.u_star[atomIdx][i] =
                    test_utilities::calculateR_Factor(calcGradientData.u_star[atomIdx][i],
                                                      refGradientData.u_star[atomIdx][i],
                                                      false);
        }

        // U_iso

        gradientR_Factor.u_iso[atomIdx] = test_utilities::calculateR_Factor(calcGradientData.u_iso[atomIdx],
                                                                            refGradientData.u_iso[atomIdx],
                                                                            false);
    }

    return test_utilities::calculateR_Factor(f_calc,f_ref,false);
}


bool TestGradientSmall::rFactorTestPassed(
    const GradientRFactorSM &rFactorData, 
    int messageLevel)
{
    size_t atomIdx, nAtoms, i;
    bool testPassed = true;
    double rFactorThreshold = 1e-6;
    nAtoms = rFactorData.atomLabel.size();
    
    
    if (messageLevel>1)
    {
        cout << endl;
        cout<< "structure name: " << rFactorData.systemLabel << endl;
    }

    for( atomIdx=0 ; atomIdx<nAtoms ; atomIdx++ )
    {

        if( (rFactorData.occupancy[atomIdx] > rFactorThreshold ) ||
            (rFactorData.u_iso[atomIdx] > rFactorThreshold) )
                testPassed = false;

        if(!rFactorData.position_cartesian[atomIdx].empty())
            for( i=0 ; i<3 ; i++ )
                if( rFactorData.position_cartesian[atomIdx][i] > rFactorThreshold)
                    testPassed = false;

        if(!rFactorData.position_fractional[atomIdx].empty())
            for (i = 0; i<3; i++)
                if (rFactorData.position_fractional[atomIdx][i] > rFactorThreshold)
                    testPassed = false;

        if(!rFactorData.u_cartesian[atomIdx].empty())
            for( i = 0 ; i<6 ; i++ )
                if (rFactorData.u_cartesian[atomIdx][i] > rFactorThreshold)
                    testPassed = false;

        if(!rFactorData.u_star[atomIdx].empty())
            for (i = 0; i<6; i++)
                if(rFactorData.u_star[atomIdx][i] > rFactorThreshold)
                    testPassed = false;

        if( messageLevel>1 )
        {
            cout<< "   " << rFactorData.atomLabel[atomIdx] << endl;

            cout<< "      occupancy " << rFactorData.occupancy[atomIdx] << endl;

            cout<< "      xyz_cart";
            for( i=0 ; i<3 ; i++ )
                cout<< " " << rFactorData.position_cartesian[atomIdx][i];
            cout<< endl;
            cout << "      xyz_fract";
            for (i = 0; i<3; i++)
                cout << " " << rFactorData.position_fractional[atomIdx][i];
            cout << endl;
            // if has U_aniso
            if( !rFactorData.u_cartesian[atomIdx].empty())
                if( (*max_element(rFactorData.u_cartesian[atomIdx].begin(), rFactorData.u_cartesian[atomIdx].end())) > -0.1)
                {
                    cout << "      U_cart";
                    for (i = 0; i<6; i++)
                        cout << " " << rFactorData.u_cartesian[atomIdx][i];
                    cout << endl;
                }
            if(!rFactorData.u_star[atomIdx].empty())
            {
                if ((*max_element(rFactorData.u_star[atomIdx].begin(), rFactorData.u_star[atomIdx].end())) > -0.1)
                {
                    cout << "      U*";
                    for (i = 0; i<6; i++)
                        cout << " " << rFactorData.u_star[atomIdx][i];
                    cout << endl;
                }
            }
            else if(rFactorData.u_iso[atomIdx]>-0.1)
                cout<< "      U_iso " << rFactorData.u_iso[atomIdx] << endl;
            

        }
    }

    return testPassed;
}

//TestGradientSmall::DerivativesProcessor::DerivativesProcessor(
//    size_t nHkl,
//    const std::vector<size_t> &nAdpsPerAtom)
//{
//    mGradientData.allocate(nHkl, nAdpsPerAtom);
//    mStructureFactors.assign(nHkl,0.0);
//}
//
//TestGradientSmall::DerivativesProcessor::~DerivativesProcessor()
//{
//}
//
//void TestGradientSmall::DerivativesProcessor::onPerHklCalculation(
//    size_t hklIdx, 
//    std::complex<double> &structureFactor, 
//    SfDerivativesAtHkl &derivatives)
//{
//    size_t atomIdx, nAtoms = mGradientData.occupancy.size();
//
//    mStructureFactors[hklIdx] = structureFactor;
//
//    for(atomIdx=0;atomIdx<nAtoms;atomIdx++)
//    {
//        mGradientData.occupancy[atomIdx][hklIdx] = derivatives.occupancyDerivatives[atomIdx];
//        
//        for(size_t i=0;i<3;i++)
//            mGradientData.position_cartesian[atomIdx][i][hklIdx] = derivatives.atomicPostionDerivatives[atomIdx][i];//.real();
//        
//        if(derivatives.adpDerivatives[atomIdx].size()==1)
//            mGradientData.u_iso[atomIdx][hklIdx] = derivatives.adpDerivatives[atomIdx][0];
//
//        if (derivatives.adpDerivatives[atomIdx].size() == 6)
//        {
//            for (size_t i = 0; i<6; i++)
//                mGradientData.u_cartesian[atomIdx][i][hklIdx] = derivatives.adpDerivatives[atomIdx][i];
//        }
//    }
//    mGradientData.hasXyzCart = true;
//    mGradientData.hasUcart = true;
//
//}
//
//void TestGradientSmall::DerivativesProcessor::clear()
//{
//}
//
//void TestGradientSmall::DerivativesProcessor::getGradientData(
//    GradientDataSM &gradientData)
//{
//    gradientData = mGradientData;
//}
//
//void TestGradientSmall::DerivativesProcessor::getStructureFactors(
//    std::vector<std::complex<double> > &scatteringFactors)
//{
//    scatteringFactors = mStructureFactors;
//}