#include "data_directory.h"

#include "Test_Multipolar.h"
//#include "Scattering/SF_Engine_Spherical_Classic.h"
#include "discamb/Scattering/HansenCoppensStructureFactorCalculator.h"
//#include "Utilities/NativeIAM_Reader.h"
#include "discamb/IO/xd_io.h"
#include "discamb/BasicUtilities/StringUtilities.h"
#include "discamb/BasicUtilities/OnError.h"
#include "discamb/CrystalStructure/ReciprocalLatticeUnitCell.h"
//#include "discamb/CrystalStructure/crystal_structure_utilities.hpp"

#include "discamb/IO/NativeIAM_Reader.h"
#include "discamb/IO/hkl_io.h"

#include "test_utilities.h"

#include <iostream>
#include <fstream>
#include <map>
#include <iomanip>

using namespace std;
using namespace discamb;


TestMultipolar::TestMultipolar()
{
    mRFactorThreshold = 0.001;
    mEngineId = "CPU";
    mN_Threads = 1;
    /*
    std::string mDataDirectory;
    bool mOneDirectory;
    */
}

TestMultipolar::~TestMultipolar()
{
}

bool TestMultipolar::run_test(
    std::string &message, 
    const std::vector<std::string> &arguments)
{
    if(arguments.size()==1)
    {
        mDataDirectory = arguments[0];
        return run_test(message, false);
    }
    return run_test(message, true);
}

bool TestMultipolar::run_test(
    std::string &message)
{
    return run_test(message,true);
}

bool TestMultipolar::run_test(
    std::string &message, 
    bool useDefaultPaths)
{
    message.clear();
    Crystal crystal;
    string prefix,xdStructureDir, structureFactorsDir;

    if(useDefaultPaths)
    {
        prefix = correcness_test_1_data_directory + string("/");
        xdStructureDir = prefix + string("structure_data/xd/");
        structureFactorsDir = prefix + string("structure_factors/HC/");
    }
    else
    {
        if(mDataDirectory.empty())
            xdStructureDir = prefix = structureFactorsDir = mDataDirectory;
        else
            xdStructureDir = prefix = structureFactorsDir = mDataDirectory + string("/");
    }

    string testInputFilePath = prefix + string("test_multipolar.inp");
    ifstream in(testInputFilePath.c_str());

    if(!in.good())
    {
        on_error::throwException(string("can not read Multipolar test input file: ")
                                 + testInputFilePath, __FILE__, __LINE__);
    }

    size_t n_entries, i;
    string s, xd_mas_file, xd_inp_file, hkl_file;
    double r_factor;
    bool passed_test, print_diff;
    string engineTypeAsString;
    
    in >> n_entries >> s >> s >> s;
    message.clear();

    print_diff = (s == string("yes"));

    in>> s >> mN_Threads >> s >> engineTypeAsString;

    mEngineId = engineTypeAsString;

    passed_test = true;

    for (i = 0; i<n_entries; ++i)
    {
        in >> xd_mas_file >> xd_inp_file >> hkl_file;
        if(mMessageLevel>1)
            cout << "processing " << xd_mas_file << endl;

        r_factor = calc_R( xdStructureDir + xd_mas_file, 
                           xdStructureDir + xd_inp_file, 
                           structureFactorsDir + hkl_file,
                           print_diff);

        passed_test = passed_test && (r_factor < mRFactorThreshold);

        if (mMessageLevel>1)
            cout<< "R factor for structure from " << xd_mas_file << " : "  << r_factor << endl;

    }
    return passed_test;
}



double TestMultipolar::calc_R(
    const std::string &xd_mas_file, 
    const std::string &xd_inp_file, 
    const std::string &hkl_file,
    bool print_diff)
{
    HC_ModelParameters hc_parameters;
    Crystal crystal;
    vector<Vector3i> hkl;
    vector<complex<double> > ref_f, f, dTarg_df;
    vector<bool> countAtomContribution;
    vector<string> words;
    vector<XdLocalCoordinateSystem> lcs;
    vector<Matrix3d> lcsMatrices;

    xd_io::read(xd_mas_file, xd_inp_file,hc_parameters,crystal, lcs, false);
    lcsMatrices.resize(lcs.size());
    for(size_t i=0;i< lcs.size();i++)
        lcs[i].calculate(lcsMatrices[i], crystal);

    string_utilities::split(xd_mas_file,words,'.');
    string outStructure = words[0] + ".str";

    //NativeIAM_Reader reader;
    //reader.write(outStructure,crystal);

    // calc iam reflection intensities

    countAtomContribution.assign(crystal.atoms.size(),true);
    hkl_io::readReflectionFile(hkl_file, hkl, ref_f);
    HansenCoppensStructureFactorCalculator calculator(crystal, hc_parameters);
    calculator.setN_Threads(mN_Threads);
    calculator.setCalculationMode(mEngineId);
    calculator.calculateStructureFactors(crystal.atoms, lcsMatrices, hkl, f, countAtomContribution);

    if(print_diff)
        printDifference(xd_mas_file, hkl, f, ref_f);

    return test_utilities::calculateR_Factor(f, ref_f, false);

}

void TestMultipolar::printDifference(
    const std::string &xd_mas_file_name, 
    std::vector<Vector3i> &hkl,
    std::vector<std::complex<double> > &ref_f, 
    std::vector<std::complex<double> > &f)
{
    ofstream out((xd_mas_file_name+string(".diff")).c_str());
    double d;
    int i,n = hkl.size();

    out<< "h     k     l   ref_f.re    calc_f.re     ref_f.im    calc_f.im    diff.re    diff.im     %diff";

    for(i=0;i<n;i++)
    {
        for(int j=0;j<3;j++)
            out<< setw(6) << hkl[i][j];
        
        out << setw(14) << fixed << setprecision(4) << ref_f[i].real();
        out << setw(14) << fixed << setprecision(4) << f[i].real();
        out << setw(14) << fixed << setprecision(4) << ref_f[i].imag();
        out << setw(14) << fixed << setprecision(4) << f[i].imag();
        out << setw(14) << fixed << setprecision(4) << ref_f[i].real() -f[i].real();
        out << setw(14) << fixed << setprecision(4) << ref_f[i].imag() - f[i].imag();
        if(norm(ref_f[i])<0.0001)
            d=0;
        else
            d=sqrt(norm(ref_f[i]-f[i])/norm(ref_f[i]));
        out << setw(14) << fixed << setprecision(4) << d << endl;
    }
    out.close();
}


