#include "data_directory.h"

#include "Test_MultipolarAsIAM.h"
#include "discamb/Scattering/HansenCoppensStructureFactorCalculator.h"
#include "discamb/IO/xd_io.h"
#include "discamb/IO/NativeIAM_Reader.h"
#include "discamb/BasicUtilities/StringUtilities.h"
#include "discamb/BasicUtilities/OnError.h"
#include "discamb/CrystalStructure/ReciprocalLatticeUnitCell.h"
//#include "discamb/CrystalStructure/crystal_structure_utilities.hpp"


#include "discamb/IO/hkl_io.h"

#include "test_utilities.h"

#include <iostream>
#include <fstream>
#include <map>
#include <iomanip>

using namespace std;
using namespace discamb;


TestMultipolarAsIAM::TestMultipolarAsIAM()
{
    mRFactorThreshold = 0.0001;
    mN_Threads = 1;
}

TestMultipolarAsIAM::~TestMultipolarAsIAM()
{
}

bool TestMultipolarAsIAM::run_test(
    std::string &message)
{
    Crystal crystal;
    string prefix = correcness_test_1_data_directory + string("/");
    ifstream in((prefix + string("test_multipolar_as_iam.inp")).c_str());

    if(!in.good())
        on_error::throwException("can not read test_multipolar_as_iam.inp",__FILE__,__LINE__);

    size_t n_entries, i;
    string s, xd_mas_file, xd_inp_file, hkl_file, structure_file;
    double r_factor;
    bool passed_test, print_diff;
    
    // #entries print diff yes/no n_cores #cores 
    in >> n_entries >> s >> s >> s;
    message.clear();

    print_diff = (s == string("yes"));

    in >> s >> mN_Threads;

    passed_test = true;
    string dirXdStructure, dirNativeStructure, dirScatterinFactors;
    dirXdStructure = prefix + string("structure_data/xd_nH/"); 
    dirNativeStructure = prefix + string("structure_data/iam/");
    dirScatterinFactors = prefix + string("structure_factors/iam/");

    for (i = 0; i<n_entries; ++i)
    {
        in >> xd_mas_file >> xd_inp_file >> structure_file >> hkl_file;
        if(mMessageLevel>0)
            cout << "processing " << xd_mas_file << endl;
        r_factor = calc_R(dirXdStructure + xd_mas_file,
                          dirXdStructure + xd_inp_file, 
                          dirNativeStructure + structure_file,
                          dirScatterinFactors + hkl_file, print_diff);
        passed_test = passed_test && (r_factor < mRFactorThreshold);

        if(mMessageLevel>1)
            cout<< "R factor for structure from " << xd_mas_file << " : " << r_factor << endl;

    }
    return passed_test;
}



double TestMultipolarAsIAM::calc_R(
    const std::string &xd_mas_file,
    const std::string &xd_inp_file,
    const std::string &structureFile,
    const std::string &hkl_file,
    bool print_diff)
{
    HC_ModelParameters hc_parameters;
    Crystal crystal;
    vector<Vector3i> hkl;
    vector<complex<double> > ref_f, f, dTarg_df;
    vector<bool> countAtomContribution;
    vector<string> words;
    vector<XdLocalCoordinateSystem> lcs;
    vector<Matrix3d> lcsMatrices;
    NativeIAM_Reader iamReader;
    ScatteringParameters scatteringParameters;
    
    
    xd_io::read(xd_mas_file, xd_inp_file, hc_parameters, crystal, lcs, false);
    iamReader.read(structureFile, crystal, scatteringParameters);

    if(!test_utilities::setAnomalousDispersionParametersInMultipolarModel(hc_parameters,scatteringParameters))
        on_error::throwException(string("wfnType should correspond to IAM type, invalid test set containing file '") + structureFile + string("'"),
                                 __FILE__, __LINE__);

    lcsMatrices.resize(lcs.size());
    for (size_t i = 0; i< lcs.size(); i++)
        lcs[i].calculate(lcsMatrices[i], crystal);

    string_utilities::split(xd_mas_file, words, '.');
    string outStructure = words[0] + ".str";

    // calc iam reflection intensities

    countAtomContribution.assign(crystal.atoms.size(), true);
    hkl_io::readReflectionFile(hkl_file, hkl, ref_f);
    HansenCoppensStructureFactorCalculator calculator(crystal, hc_parameters);
    
    calculator.setN_Threads(mN_Threads);
    //calculator..setEngineType(CPU_IAM);
    calculator.useCPU_IAM();
    calculator.calculateStructureFactors(crystal.atoms, lcsMatrices, hkl, f, countAtomContribution);

    if (print_diff)
        printDifference(xd_mas_file, hkl, f, ref_f);

    return test_utilities::calculateR_Factor(f, ref_f, false);

}

void TestMultipolarAsIAM::printDifference(
    const std::string &xd_mas_file_name,
    std::vector<Vector3i> &hkl,
    std::vector<std::complex<double> > &ref_f,
    std::vector<std::complex<double> > &f)
{
    ofstream out((xd_mas_file_name + string(".diff")).c_str());
    double d;
    int i, n = hkl.size();

    out << "h     k     l   ref_f.re    calc_f.re     ref_f.im    calc_f.im    diff.re    diff.im     %diff";

    for (i = 0; i<n; i++)
    {
        for (int j = 0; j<3; j++)
            out << setw(6) << hkl[i][j];

        out << setw(14) << fixed << setprecision(4) << ref_f[i].real();
        out << setw(14) << fixed << setprecision(4) << f[i].real();
        out << setw(14) << fixed << setprecision(4) << ref_f[i].imag();
        out << setw(14) << fixed << setprecision(4) << f[i].imag();
        out << setw(14) << fixed << setprecision(4) << ref_f[i].real() - f[i].real();
        out << setw(14) << fixed << setprecision(4) << ref_f[i].imag() - f[i].imag();
        if (norm(ref_f[i])<0.0001)
            d = 0;
        else
            d = sqrt(norm(ref_f[i] - f[i]) / norm(ref_f[i]));
        out << setw(14) << fixed << setprecision(4) << d << endl;
    }
    out.close();
}


