#include "data_directory.h"

#include "Test_MultipolarConversion.h"
#include "Test_GradientLarge.h"
#include "Test_GradientSmall.h"
#include "GradientDataCollector.h"

#include "discamb/Scattering/HansenCoppensStructureFactorCalculator.h"
#include "discamb/IO/xd_io.h"
#include "discamb/IO/NativeIAM_Reader.h"
#include "discamb/BasicUtilities/StringUtilities.h"
#include "discamb/BasicUtilities/OnError.h"
#include "discamb/CrystalStructure/ReciprocalLatticeUnitCell.h"
//#include "discamb/CrystalStructure/crystal_structure_utilities.hpp"


#include "discamb/IO/hkl_io.h"

#include "test_utilities.h"

#include <iostream>
#include <fstream>
#include <map>
#include <iomanip>

using namespace std;
using namespace discamb;


TestMultipolarConversion::TestMultipolarConversion()
{
}

TestMultipolarConversion::~TestMultipolarConversion()
{
}

bool TestMultipolarConversion::run_test(
    std::string &message)
{
    Crystal crystal;
    string prefix = correcness_test_1_data_directory + string("/");
    ifstream in((prefix + string("test_multipolar_conversion.inp")).c_str());


    if (!in.good())
        on_error::throwException("can not read test_multipolar_conversion.inp", __FILE__, __LINE__);

    size_t n_entries, i;
    string s, xd_mas_file, xd_inp_file, hkl_file, structure_file_R_frac_U_cif;
    string structure_file_R_cart_U_cart, structure_file_R_cart_U_star, gradLargeFile, gradSmallFile, dTarget_dF_File;
    bool passed_test, entryTestPassed;
    vector<complex<double> > dTarget_dF;

    // #entries 
    in>> n_entries >> s >> mN_Threads;
    
    message.clear();


    passed_test = true;

    string dirXdStructure, dirIamStructure, dirScatterinFactors, dirGradLarge, dirGradSmall;
    dirXdStructure = prefix + string("structure_data/xd_nH/");
    dirIamStructure = prefix + string("structure_data/iam/");
    dirGradSmall = prefix + string("gradient_data/small_molecule/iam/");
    

    for (i = 0; i<n_entries; ++i)
    {
        in >> xd_mas_file >> xd_inp_file >> structure_file_R_frac_U_cif >> structure_file_R_cart_U_cart 
           >> structure_file_R_cart_U_star >> gradSmallFile;
        if (mMessageLevel>0)
            cout << "processing " << xd_mas_file << endl;

        entryTestPassed = testStructure(dirXdStructure + xd_mas_file,
                                        dirXdStructure + xd_inp_file,
                                        dirIamStructure + structure_file_R_frac_U_cif,
                                        dirIamStructure + structure_file_R_cart_U_cart,
                                        dirIamStructure + structure_file_R_cart_U_star,
                                        dirGradSmall + gradSmallFile);

        passed_test = passed_test && entryTestPassed;

        if (mMessageLevel>1)
        {
            cout << "test for structure from " << xd_mas_file << " : ";
            if(entryTestPassed) 
                cout<< "passed" << endl;
            else 
                cout<< "failed" << endl;
        }

    }
    return passed_test;
}

bool TestMultipolarConversion::testStructure(
    const std::string &xd_mas,
    const std::string &xd_inp,
    const std::string &structure_file_R_frac_U_cif,
    const std::string &structure_file_R_cart_U_cart,
    const std::string &structure_file_R_cart_U_star,
    const std::string &gradSmallFile)
{
    Crystal crystal;
    vector<complex<double> > f;
    ScatteringParameters scatteringParameters;
    std::vector<std::complex<double> > dTarget_dF(1,1.0);
    NativeIAM_Reader reader;
    std::vector<TargetFunctionAtomicParamDerivatives> dTarg_dParam;
    size_t hklIdx, nHkl;
    HC_ModelParameters hc_parameters;
    vector< XdLocalCoordinateSystem > lcs;
    vector< complex< double > > ref_f_large, f_large, f_small, ref_f_small, f1, f2, f3;
    vector< Vector3i > hklLarge, hklSmall;
    //vector< TestGradientLarge::GradientData > gradientDataLarge;
    GradientDataSM gradientDataSM_frac_Ustar, gradientDataSM_cart_Ucart, gradientDataSM_cart_Ucif,
                   gradientDataLM_frac_Ustar, gradientDataLM_cart_Ucart, gradientDataLM_cart_Ucif;
    GradientDataSM_AllConventions refDataSM;
    //string mainDir = correcness_test_1_data_directory;

    // read reference data

    TestGradientSmall::readGradientAndHkl(gradSmallFile, ref_f_small, refDataSM, hklSmall);
    
    // read structural data

    xd_io::read(xd_mas, xd_inp, hc_parameters, crystal, lcs, false);

    reader.read(structure_file_R_frac_U_cif, crystal, scatteringParameters);

    vector<bool> countAtomContribution(crystal.atoms.size(), true);


    if (!test_utilities::setAnomalousDispersionParametersInMultipolarModel(hc_parameters, scatteringParameters))
        on_error::throwException(string("wfnType should correspond to IAM type, invalid test set containing file '") + xd_mas + string("'"),
                                 __FILE__, __LINE__);

    // preapare local coordinates

    vector<Matrix3d> lcsMatrices;
    lcsMatrices.resize(lcs.size());
    for (size_t i = 0; i< lcs.size(); i++)
        lcs[i].calculate(lcsMatrices[i], crystal);

    // set structure factors calculator
    HansenCoppensStructureFactorCalculator hcCalculator(crystal, hc_parameters);
    //hcCalculator.setEngineType(CPU_IAM);
    hcCalculator.useCPU_IAM();
    hcCalculator.setN_Threads(1);

    
    // set nAdpsPerAtom
    
    size_t nAtoms = crystal.atoms.size();
    vector<size_t> nAdpsPerAtom(nAtoms);
    for(size_t i=0; i< nAtoms; i++)
        nAdpsPerAtom[i] = crystal.atoms[i].adp.size();

    nHkl = hklSmall.size();
    gradientDataLM_frac_Ustar.allocate(nHkl, nAdpsPerAtom);
    gradientDataLM_cart_Ucart.allocate(nHkl, nAdpsPerAtom); 
    gradientDataLM_cart_Ucif.allocate(nHkl, nAdpsPerAtom);

    //#################################################
    //
    //     in R_frac_U_cif out R frac U star
    //
    //#################################################

    crystal.xyzCoordinateSystem = structural_parameters_convention::fractional;
    crystal.adpConvention = structural_parameters_convention::U_cif;

    hcCalculator.setDerivativesConvention(structural_parameters_convention::fractional, structural_parameters_convention::U_star);
    
    hcCalculator.setModel(crystal, hc_parameters);

    // version for small molecules
    
    
    GradientDataCollector gradientDataCollector(nHkl , nAdpsPerAtom);
    hcCalculator.calculateStructureFactorsAndDerivatives(crystal.atoms, lcsMatrices, hklSmall, gradientDataCollector);
    gradientDataCollector.getGradientData(gradientDataSM_frac_Ustar);
    gradientDataCollector.getStructureFactors(f1);
    
    // version for large molecules
    
    hklLarge.resize(1);
    for( hklIdx=0 ; hklIdx < nHkl ; hklIdx++ )
    {
        hklLarge[0] = hklSmall[hklIdx];
        hcCalculator.calculateStructureFactorsAndDerivatives(crystal.atoms, lcsMatrices, hklLarge, f, dTarg_dParam,
                                                             dTarget_dF, countAtomContribution);
        addGradientInfo(dTarg_dParam, gradientDataLM_frac_Ustar, hklIdx);
    }

    //########################################################
    //
    //     in structure_file_R_cart_U_cart out R cart U cart
    //
    //########################################################

    reader.read(structure_file_R_cart_U_cart, crystal, scatteringParameters);

    crystal.xyzCoordinateSystem = structural_parameters_convention::cartesian;
    crystal.adpConvention = structural_parameters_convention::U_cart;

    hcCalculator.setDerivativesConvention(structural_parameters_convention::cartesian, structural_parameters_convention::U_cart);
    hcCalculator.setModel(crystal, hc_parameters);
    // version for small molecules

    nHkl = hklSmall.size();
    hcCalculator.calculateStructureFactorsAndDerivatives(crystal.atoms, lcsMatrices, hklSmall, gradientDataCollector);
    gradientDataCollector.getGradientData(gradientDataSM_cart_Ucart);
    gradientDataCollector.getStructureFactors(f2);

    // version for large molecules

    for (hklIdx = 0; hklIdx < nHkl; hklIdx++)
    {
        hklLarge[0] = hklSmall[hklIdx];
        hcCalculator.calculateStructureFactorsAndDerivatives(crystal.atoms, lcsMatrices, hklLarge, f, dTarg_dParam,
                                                             dTarget_dF, countAtomContribution);
        addGradientInfo(dTarg_dParam, gradientDataLM_cart_Ucart, hklIdx);
    }

    //#######################################################
    //
    //     in structure_file_R_cart_U_star out R cart U cif
    //
    //#######################################################

    reader.read(structure_file_R_cart_U_star, crystal, scatteringParameters);

    crystal.xyzCoordinateSystem = structural_parameters_convention::cartesian;
    crystal.adpConvention = structural_parameters_convention::U_star;

    hcCalculator.setDerivativesConvention(structural_parameters_convention::cartesian, structural_parameters_convention::U_cif);
    hcCalculator.setModel(crystal, hc_parameters);
    // version for small molecules

    nHkl = hklSmall.size();
    hcCalculator.calculateStructureFactorsAndDerivatives(crystal.atoms, lcsMatrices, hklSmall, gradientDataCollector);
    gradientDataCollector.getGradientData(gradientDataSM_cart_Ucif);
    gradientDataCollector.getStructureFactors(f3);

    // version for large molecules

    for (hklIdx = 0; hklIdx < nHkl; hklIdx++)
    {
        hklLarge[0] = hklSmall[hklIdx];
        hcCalculator.calculateStructureFactorsAndDerivatives(crystal.atoms, lcsMatrices, hklLarge, f, dTarg_dParam,
                                                             dTarget_dF, countAtomContribution);
        addGradientInfo(dTarg_dParam, gradientDataLM_cart_Ucif, hklIdx);
    }

    //######################################################
    //
    //     compare with reference data
    //
    //######################################################

    return compare(refDataSM, gradientDataSM_frac_Ustar, gradientDataSM_cart_Ucart, gradientDataSM_cart_Ucif, 
                   gradientDataLM_frac_Ustar, gradientDataLM_cart_Ucart, gradientDataLM_cart_Ucif);
}

void TestMultipolarConversion::addGradientInfo(
    const  std::vector<TargetFunctionAtomicParamDerivatives> &dTarg_dParam, 
    GradientDataSM &dataSM,
    size_t hklIdx)
{
    size_t i, atomIdx, nAtoms = dTarg_dParam.size();
    

    for(atomIdx=0; atomIdx<nAtoms;atomIdx++)
    {
        if(dTarg_dParam[atomIdx].adp_derivatives.size()==6)
            for(i=0; i<6; i++)
                dataSM.dU_aniso[atomIdx][i][hklIdx] = dTarg_dParam[atomIdx].adp_derivatives[i];
        if (dTarg_dParam[atomIdx].adp_derivatives.size() == 1)
            dataSM.dU_iso[atomIdx][hklIdx] = dTarg_dParam[atomIdx].adp_derivatives[0];
        for(i=0; i<3;i++)
            dataSM.dPosition[atomIdx][i][hklIdx] = dTarg_dParam[atomIdx].atomic_position_derivatives[i];
        dataSM.dOccupancy[atomIdx][hklIdx] = dTarg_dParam[atomIdx].occupancy_derivatives;
    }
}



bool TestMultipolarConversion::compare(
    const GradientDataSM_AllConventions &dataAll,
    const GradientDataSM &gradientDataSM_frac_Ustar,
    const GradientDataSM &gradientDataSM_cart_Ucart,
    const GradientDataSM &gradientDataSM_cart_Ucif,
    const GradientDataSM &gradientDataLM_frac_Ustar,
    const GradientDataSM &gradientDataLM_cart_Ucart,
    const GradientDataSM &gradientDataLM_cart_Ucif)
{
    size_t i, atomIdx, nAtoms = dataAll.occupancy.size();

    for(atomIdx=0; atomIdx<nAtoms; atomIdx++)
    {
        // occupancy
        if( !compare(dataAll.occupancy[atomIdx], gradientDataSM_cart_Ucart.dOccupancy[atomIdx]))
            return false;
        if (!compare(dataAll.occupancy[atomIdx], gradientDataSM_cart_Ucif.dOccupancy[atomIdx]))
            return false;
        if (!compare(dataAll.occupancy[atomIdx], gradientDataSM_frac_Ustar.dOccupancy[atomIdx]))
            return false;
        if (!compare(dataAll.occupancy[atomIdx], gradientDataLM_cart_Ucart.dOccupancy[atomIdx], true))
            return false;
        if (!compare(dataAll.occupancy[atomIdx], gradientDataLM_cart_Ucif.dOccupancy[atomIdx], true))
            return false;
        if (!compare(dataAll.occupancy[atomIdx], gradientDataLM_frac_Ustar.dOccupancy[atomIdx], true))
            return false;

        // xyz cart
        for(i=0;i<3;i++) // lop over position vector components (x,y,z)
        {
            if (!compare(dataAll.position_cartesian[atomIdx][i], gradientDataSM_cart_Ucart.dPosition[atomIdx][i]))
                return false;
            if (!compare(dataAll.position_cartesian[atomIdx][i], gradientDataSM_cart_Ucif.dPosition[atomIdx][i]))
                return false;
            if (!compare(dataAll.position_cartesian[atomIdx][i], gradientDataLM_cart_Ucart.dPosition[atomIdx][i], true))
                return false;
            if (!compare(dataAll.position_cartesian[atomIdx][i], gradientDataLM_cart_Ucif.dPosition[atomIdx][i], true))
                return false;
        }

        // xyz frac
        for (i = 0; i<3; i++) // lop over position vector components (x,y,z)
        {
            if (!compare(dataAll.position_fractional[atomIdx][i], gradientDataSM_frac_Ustar.dPosition[atomIdx][i]))
                return false;
            if (!compare(dataAll.position_fractional[atomIdx][i], gradientDataLM_frac_Ustar.dPosition[atomIdx][i], true))
                return false;
        }

        // U_iso
        if(!dataAll.u_iso[atomIdx].empty())
        {
            if (!compare(dataAll.u_iso[atomIdx], gradientDataSM_cart_Ucart.dU_iso[atomIdx]))
                return false;
            if (!compare(dataAll.u_iso[atomIdx], gradientDataSM_cart_Ucif.dU_iso[atomIdx]))
                return false;
            if (!compare(dataAll.u_iso[atomIdx], gradientDataSM_frac_Ustar.dU_iso[atomIdx]))
                return false;
            if (!compare(dataAll.u_iso[atomIdx], gradientDataLM_cart_Ucart.dU_iso[atomIdx], true))
                return false;
            if (!compare(dataAll.u_iso[atomIdx], gradientDataLM_cart_Ucif.dU_iso[atomIdx], true))
                return false;
            if (!compare(dataAll.u_iso[atomIdx], gradientDataLM_frac_Ustar.dU_iso[atomIdx], true))
                return false;
        }

        // U_cart
        if (!dataAll.u_cartesian[atomIdx][0].empty())
        {
            for(i=0;i<6;i++)
            {
            if (!compare(dataAll.u_cartesian[atomIdx][i], gradientDataSM_cart_Ucart.dU_aniso[atomIdx][i]))
                return false;
            if (!compare(dataAll.u_cartesian[atomIdx][i], gradientDataLM_cart_Ucart.dU_aniso[atomIdx][i], true))
                return false;
            }
        }

        // U_star
        if (!dataAll.u_star[atomIdx][0].empty())
        {
            for (i = 0; i<6; i++)
            {
                if (!compare(dataAll.u_star[atomIdx][i], gradientDataSM_frac_Ustar.dU_aniso[atomIdx][i]))
                    return false;
                if (!compare(dataAll.u_star[atomIdx][i], gradientDataLM_frac_Ustar.dU_aniso[atomIdx][i], true))
                    return false;
            }
        }

        // U_cif
        /*
        if (!dataAll.u_cif[atomIdx].empty())
        {
            for (i = 0; i<6; i++)
            {
                if (!compare(dataAll.u_cif[atomIdx][i], gradientDataSM_cart_Ucif.dU_aniso[atomIdx][i]))
                    return false;
                if (!compare(dataAll.u_cif[atomIdx][i], gradientDataSM_cart_Ucif.dU_aniso[atomIdx][i], true))
                    return false;
            }
        }
        */
    }
    return true;
}


bool TestMultipolarConversion::compare(
    const std::vector<std::complex<double> > &ref, 
    const std::vector<std::complex<double> > &calc,
    bool onlyRealPart)
{
    if(ref.empty())
        return true;
    if(ref.size() != calc.size())
        on_error::throwException("BUG, incompatible data set sizes", __FILE__,__LINE__);
    double maxReal, maxIm, absRef, absCalc, r, c, error, diff;
    maxReal = 0;
    maxIm = 0;
    size_t i, n = ref.size();
    double minValueThreshold = 1e-6;
    double errorThreshold = 5e-4;
    
    for(i=0;i<n;i++)
    {
        if(fabs(ref[i].real())>maxReal)
            maxReal = fabs(ref[i].real());
        if (fabs(ref[i].imag())>maxIm)
            maxIm = fabs(ref[i].imag());
    }

    for (i = 0; i<n; i++)
    {
        r = ref[i].real();
        c = calc[i].real();
        diff = r - c;
        absRef = fabs(r);
        absCalc = fabs(c);

        if( absRef > 1e-5*maxReal)
            if( absRef > minValueThreshold || absCalc>minValueThreshold)
            {

                absRef > absCalc ? error = diff*diff/(absRef*absRef) : error = diff*diff / (absCalc*absCalc);
                if(error > errorThreshold)
                    return false;
            }

        if(!onlyRealPart)
        {
            r = ref[i].imag();
            c = calc[i].imag();
            diff = r - c;
            absRef = fabs(r);
            absCalc = fabs(c);


            if (absRef > 1e-5*maxIm)
                if (absRef > minValueThreshold || absCalc>minValueThreshold)
                {

                    absRef > absCalc ? error = diff*diff / (absRef*absRef) : error = diff*diff / (absCalc*absCalc);
                    if (error > errorThreshold)
                        return false;
                }
        }
    }
    return true;
}

