#include "Test_MultipolarGradientLarge.h"
#include "test_utilities.h"
#include "data_directory.h"

#include "discamb/MathUtilities/Vector3.h"

#include "discamb/BasicUtilities/OnError.h"
#include "discamb/BasicUtilities/StringUtilities.h"
#include "discamb/CrystalStructure/ReciprocalLatticeUnitCell.h"
#include "discamb/Scattering/HansenCoppensStructureFactorCalculator.h"

#include "discamb/IO/NativeIAM_Reader.h"
#include "discamb/IO/xd_io.h"


#include <fstream>
#include <iomanip>

#ifdef _OPENMP
#include <omp.h>
#endif

using namespace std;
using namespace discamb;





TestMultipolarGradientLarge::TestMultipolarGradientLarge()
{
    mN_Cores=0;
}

TestMultipolarGradientLarge::~TestMultipolarGradientLarge()
{
}

string TestMultipolarGradientLarge::gradientFilesDir()
const
{
    return correcness_test_1_data_directory + string("/gradient_data/large_molecule/HC");
}


void TestMultipolarGradientLarge::readStructureAndCalculateDerivatives(
    bool customDir,
    const std::string &customDirPath,
    const std::vector<std::string> &structureFileNames,
    const std::vector<Vector3i> &hkl,
    const vector<complex<double> > &dTarget_dF,
    Crystal &crystal,
    std::vector<TargetFunctionAtomicParamDerivatives> &dTarg_dParam) 
{
    vector<complex<double> > f;
    ScatteringParameters scatteringParameters;
    NativeIAM_Reader reader;
    Crystal crystal2;
    string dataDir = correcness_test_1_data_directory;
    string iamStructureDir = makePrefix(customDir, customDirPath, dataDir + "/structure_data/iam");
    string xdStructureDir = makePrefix(customDir, customDirPath, dataDir + "/structure_data/xd");


    HC_ModelParameters hc_parameters;
    vector<XdLocalCoordinateSystem> lcs;
    xd_io::read(xdStructureDir+structureFileNames[0], 
                    xdStructureDir+structureFileNames[1], 
                    hc_parameters, crystal, lcs, false);
    
    reader.read(iamStructureDir + structureFileNames[2], crystal, scatteringParameters);


    vector<bool> countAtomContribution(crystal.atoms.size(), true);

    
    if(!test_utilities::setAnomalousDispersionParametersInMultipolarModel(hc_parameters, scatteringParameters))
        on_error::throwException(string("wfnType should correspond to IAM type, invalid test set containing file '") + structureFileNames[0] + string("'"),
        __FILE__, __LINE__);


    vector<Matrix3d> lcsMatrices;


    lcsMatrices.resize(lcs.size());
    for (size_t i = 0; i< lcs.size(); i++)
        lcs[i].calculate(lcsMatrices[i], crystal);

    HansenCoppensStructureFactorCalculator hcCalculator(crystal, hc_parameters);
    hcCalculator.setDerivativesConvention(structural_parameters_convention::cartesian, structural_parameters_convention::U_cart);
    hcCalculator.setCalculationMode(mEngineId);
    hcCalculator.setN_Threads(mN_Cores);
    hcCalculator.calculateStructureFactorsAndDerivatives(crystal.atoms, lcsMatrices, hkl, f, dTarg_dParam, dTarget_dF,countAtomContribution);
    if(mWrite)
        writeResults(structureFileNames[0], crystal.atoms, hkl, f, dTarg_dParam);
        //writeResults(xdStructureDir + structureFileNames[0], crystal.atoms, hkl, f, dTarg_dParam);
}

void TestMultipolarGradientLarge::readInputFile(
    const std::string &inputFileName,
    std::string &dTarget_dF_file,
    std::vector<std::vector<string> > &structureFiles,
    std::vector<std::string> &hklFiles,
    std::vector<std::string> &gradientFiles) 
{

    ifstream in(inputFileName.c_str());
    string s, line;
    int nCores,nAvailableThreads;
    vector<string> words;
    string engineIdAsString;

    if (!in.good())
        on_error::throwException(string("can not open input file: '") + inputFileName +
                                 string("' when testing gradients"), __FILE__, __LINE__);
    in >> s >> s >> s >> nCores;
    in >> s >> s >> engineIdAsString;

    mEngineId = engineIdAsString;

#ifdef _OPENMP    
    if(nCores<0)
    {
        nAvailableThreads = omp_get_max_threads();
        if(nAvailableThreads+nCores>0)
            mN_Cores = nAvailableThreads + nCores;
        else
            mN_Cores = 1;
    }
    else 
        mN_Cores = nCores;
#else
    mN_Cores = 1;
#endif
    in >> s >> s >> dTarget_dF_file;
    in >> s >> s >> s;
    getline(in, line);

    size_t nDataSets = atoi(s.c_str());
    
    structureFiles.clear();
    structureFiles.resize(nDataSets);
    hklFiles.resize(nDataSets); 
    gradientFiles.resize(nDataSets); 

    for (size_t i = 0; i<nDataSets; i++)
    {
        if (!in.good())
            on_error::throwException(string("unexpected end of input file: '") + inputFileName +
                                     string("' when testing gradients"), __FILE__, __LINE__);
        getline(in, line);
        string_utilities::split(line, words);
        if (words.size() != 5)
            on_error::throwException(string("invalid format of input file: '") + inputFileName +
                                     string("' when testing gradients"), __FILE__, __LINE__);
        structureFiles[i].push_back(words[0]);
        structureFiles[i].push_back(words[1]);
        structureFiles[i].push_back(words[2]);
        hklFiles[i] = words[3];
        gradientFiles[i] = words[4];
    }

    in.close();

}

//void TestMultipolarGradientLarge::writeResults(
//    const std::string &xdMasFileName,
//    const std::vector<AtomInCrystal> &atoms,
//    const std::vector<Vector3i> &hkl,
//    const std::vector<std::complex<double> > &f,
//    std::vector<TargetFunctionAtomicParamDerivatives> &dTarg_dParam)
//{
//    // filename
//    string fileNameCore,fileName = xdMasFileName;
//
//    if (fileName.find_last_of('\\') != string::npos)
//        fileName = fileName.substr(fileName.find_last_of('\\') + 1);
//    if (fileName.find_last_of('/') != string::npos)
//        fileName = fileName.substr(fileName.find_last_of('/') + 1);
//
//    if(fileName.size()>7)
//    {
//        if(fileName.substr(fileName.size()-7)==string("_xd.mas"))
//            fileNameCore = fileName.substr(0, fileName.size() - 7);
//        else
//            fileNameCore = fileName.substr(0, fileName.size() - 4);
//    }
//    else
//        fileNameCore = fileName.substr(0, fileName.size() - 4);
//
//    //fileNameCore = fileNameCore.substr()
//    // write structure factors
//    ofstream outSF((fileNameCore+string(".sf")).c_str());
//    size_t hklIndex, nHkl = hkl.size();
//
//    if(!outSF.good())
//        on_error::throwException("can not write structure factor file for saving MultipolarGradientLarge results",__FILE__,__LINE__);
//
//    outSF<< "number of reflections " << nHkl << endl;
//
//    outSF<< setprecision(12);
//
//    for(hklIndex = 0 ; hklIndex<nHkl ; hklIndex++)
//        outSF<< hkl[hklIndex].x << " " << hkl[hklIndex].y << " " << hkl[hklIndex].z
//             << " " << f[hklIndex].real() << " " << f[hklIndex].real() << endl;
//
//    outSF.close();
//
//    // write gradient data
//
//    ofstream out((fileNameCore + string(".grd")).c_str());
//    size_t atomIdx, nAtoms = atoms.size();
//    out<< setprecision(12);
//
//    for( atomIdx = 0 ; atomIdx < nAtoms ; atomIdx++ )
//    {
//        out<< atoms[atomIdx].label << endl;
//        out<< "   occ " << dTarg_dParam[atomIdx].occupancy_derivatives << endl;
//        out<< "   xyz_Cart ";
//        for(int i=0;i<3;i++)
//            out<< dTarg_dParam[atomIdx].atomic_position_derivatives[i] << " ";
//        out<< endl;
//        
//        if(dTarg_dParam[atomIdx].adp_derivatives.size()==6)
//        {
//            out << "   U_cart";
//            for (int i = 0; i<6; i++)
//                out << " " << dTarg_dParam[atomIdx].adp_derivatives[i];
//            out << endl;
//        }
//        if (dTarg_dParam[atomIdx].adp_derivatives.size() == 1)
//            out << "   U_iso " << dTarg_dParam[atomIdx].adp_derivatives[0] << endl;
//    }
//    out.close();
//    
//}
