#include "discamb/IO/xd_io.h"
#include "discamb/IO/hkl_io.h"
#include "discamb/BasicUtilities/Exception.h"
#include "discamb/BasicUtilities/StringUtilities.h"
#include "discamb/BasicUtilities/OnError.h"
#include "discamb/Scattering/HansenCoppensStructureFactorCalculator.h"

#include "data_directory.h"

#include <iostream>
#include <map>

using namespace std;
using namespace discamb;


void calculationTime
(
    const char *xdMasterFileName,
    const char *xdParameterFileName,
    const vector<Vector3i> &hklIndices,
    int nThreads,
    const std::string &sfEngineIdAsString,
    double &cpuTime,
    double &wallClockTime)
{
    HC_ModelParameters hcModelParameters;
    Crystal crystal;
    vector<complex<double> > scattering_factors;
    vector<XdLocalCoordinateSystem> localCoordinateSystems;
    vector<Matrix3d> localCoordinateSystemMatrices;


    // reads informatoin on model from XD files
    xd_io::read(xdMasterFileName, xdParameterFileName, hcModelParameters, crystal, localCoordinateSystems, true);

    // set local coordinate system matrices
    size_t nAtoms = localCoordinateSystems.size();
    localCoordinateSystemMatrices.resize(nAtoms);
    for (size_t atomIdx = 0; atomIdx<nAtoms; atomIdx++)
        localCoordinateSystems[atomIdx].calculate(localCoordinateSystemMatrices[atomIdx], crystal);

    // scattering factor calculations engine
    HansenCoppensStructureFactorCalculator calculator(crystal, hcModelParameters);
    calculator.setN_Threads(nThreads);
    calculator.setCalculationMode(sfEngineIdAsString);

    calculator.calculateStructureFactors(crystal.atoms, localCoordinateSystemMatrices, hklIndices, scattering_factors);

    calculator.getCalculationTime(cpuTime, wallClockTime);

}


void getDataInfo(
    const char * inputFileName,
    string dirPrefix,
    vector<int> &nCores,
    std::string &engineId,
    vector<vector<string> > &dataFiles)
{
    string path;
    if(dirPrefix.empty())
        path = inputFileName;
    else
        path = dirPrefix + string("/") + inputFileName;
    ifstream in(path.c_str());
    vector<string> words;
    vector<string> engineTypes;
    engineTypes.push_back("CPU");
    engineTypes.push_back("GPU");
    engineTypes.push_back("IAM");
    engineTypes.push_back("CPU_ALT");
    string line;

    if(!in.good())
        on_error::throwException(string("cannot read performance test input file: '") + string(inputFileName) + string("'"), __FILE__, __LINE__);

    while (in.good())
    {
        getline(in, line);
        string_utilities::split(line, words);
        if (!words.empty())
            if (words[0][0] != '!')
            {
                if (find(engineTypes.begin(), engineTypes.end(), words[0]) != engineTypes.end())
                {
                    engineId = words[0];
                    if (words.size()>1 && engineId != string("GPU"))
                    {
                        nCores.clear();
                        for (size_t i = 1; i<words.size(); i++)
                            nCores.push_back(atoi(words[i].c_str()));
                    }
                }
                else
                {
                    dataFiles.push_back(words);
                    for(size_t i=0;i<words.size();i++)
                    {
                        if(!dirPrefix.empty())
                            dataFiles.back()[i] = dirPrefix + string("/") + dataFiles.back()[i];
                    }
                }

            }
    }
    in.close();

}

string getFileNameFromFullPath(
    const string &path)
{
    string result = path;
    if(path.find_last_of('/')!=string::npos)
        result = path.substr(path.find_last_of('/')+1);
    else if (path.find_last_of('\\') != string::npos)
        result = path.substr(path.find_last_of('\\') + 1);
    return result;
}

void runCalculations(
    vector<int> &nCores,
    std::string engineId,
    vector<vector<string> > &dataFiles)
{
    vector<string> hklFiles;
    vector<Vector3i> hklIndices;
    string masFileName, hklFileName;
    double wallClockTime, cpuTime;

    for (size_t dataIdx = 0; dataIdx < dataFiles.size(); dataIdx++)
    {
        hklFiles.assign(dataFiles[dataIdx].begin() + 2, dataFiles[dataIdx].end());

        masFileName = getFileNameFromFullPath(dataFiles[dataIdx][0]);


        for (size_t hklSetIdx = 0; hklSetIdx<hklFiles.size(); hklSetIdx++)
        {
            hkl_io::readCompactHklFile(hklFiles[hklSetIdx], hklIndices);

            hklFileName = getFileNameFromFullPath(hklFiles[hklSetIdx]);
            
            if (nCores.empty())
            {
                calculationTime(dataFiles[dataIdx][0].c_str(), dataFiles[dataIdx][1].c_str(), hklIndices, 1,
                                engineId, cpuTime, wallClockTime);
                cout << masFileName << " " << hklFileName << " 1 CPU core or GPU, wall clock time (ms) "
                     << wallClockTime << " cpu time (ms) " << cpuTime << endl;
            }
            else
                for (size_t nCoresIdx = 0; nCoresIdx<nCores.size(); nCoresIdx++)
                {
                    calculationTime(dataFiles[dataIdx][0].c_str(), dataFiles[dataIdx][1].c_str(), hklIndices, nCores[nCoresIdx],
                                    engineId, cpuTime, wallClockTime);

                    cout << masFileName << " " << hklFileName << " " << nCores[nCoresIdx] << " cores wall clock time (ms) "
                        << wallClockTime << " cpu time (ms) " << cpuTime << endl; 
                }

        }

    }

}


void run(const char * inputFileName)
{    
    vector<int> nCores;
    string engineIdAsString;
    vector<vector<string> > fileNames;

    getDataInfo(inputFileName, string(), nCores, engineIdAsString, fileNames);

    runCalculations( nCores, engineIdAsString, fileNames);       
}



void run()
{
    vector<int> nCores;
    string engineIdAsString;
    vector<vector<string> > fileNames;

    getDataInfo("performance.inp", performance_data_directory, nCores, engineIdAsString, fileNames);

    runCalculations( nCores, engineIdAsString, fileNames);
}


void run(
    const char *xdMasterFileName,
    const char *xdParameterFileName,
    const char *hklFileName,
    int hIndexColumn,
    int nThreads,
    const std::string &sfEngineIdAsString)
{
    // read reflections for which the scattering factors will be calculated
    vector<Vector3i> hklIndices;
    hkl_io::readHkl(hklFileName, hIndexColumn, hklIndices);

    double cpuTime, wallClockTime;

    calculationTime(xdMasterFileName, xdParameterFileName,
                    hklIndices, nThreads, sfEngineIdAsString, cpuTime, wallClockTime);

    cout << xdMasterFileName << " " << hklFileName << " " << nThreads << " cores wall clock time (ms) "
         << wallClockTime << " cpu time (ms) " << cpuTime << endl;

}

int main(int argc, char *argv[])
{

    try
    {
        if (argc != 7 && argc !=2 && argc != 1)
        {
            cout << "expected 0, 1 or 6 arguments:" << endl
                 << " 0 arguments - default input file (performance.inp) is used from " << performance_data_directory << endl
                 << " 1 argument - name of the input file" << endl
                 << " 6 arguments:" << endl
                 << "    (1) name of xd master file" << endl
                 << "    (2) name of xd parameter file" << endl
                 << "    (3) name of hkl file" << endl
                 << "    (4) number of column in hkl file with index h of reflection" << endl
                 << "    (5) number of cores to use" << endl
                 << "    (6) sf engine type" << endl;
            exit(0);
        }

        if(argc==7)
            run(argv[1], argv[2], argv[3], atoi(argv[4]), atoi(argv[5]), argv[6]);
        if(argc==2)
            run(argv[1]);
        if(argc==1)
            run();
    }
    catch (const Exception & e)
    {
        cout << e.what();
    }
    catch (const std::exception& e)
    {
        cout << "standard exception occured : " << e.what() << endl;
    }

    return 0;
}
